(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List.
Require Import vgtac basis myclassical.
Require Import lang_common lang_spec lang_mid lang_tgt.
Require Import model_common model_untyped model_properties_untyped.
Require Import gw_common gw_untyped.

Set Implicit Arguments. 
Global Set Bullet Behavior "Strict Subproofs".

Notation index x ctx := (withd 0 (list_index x ctx)). 

Definition getVal (x: itermvar) imports valctx r :=
  match x with
    | inl f =>
      match lookup f imports with
        | Some addr => i_ld r (o_imm addr) :: nil
        | None => nil 
      end
    | inr y =>
      let index := index y valctx in
      i_ld r (o_reg r_env) ::
      repeat (i_ld r (o_hind r o_one)) index ++
      i_ld r (o_hind r o_zero) :: nil
  end.

Definition getValNat (xn: itermvar + nat) imports valctx r :=
  match xn with
    | inl x => getVal x imports valctx r
    | inr n => i_ld r (o_imm n) :: nil
  end.

Definition getCont (xk: contvar) contctx :=
  let index := index xk contctx in
  i_opr Minus r_sp (o_reg r_sp) (o_imm (2 * S index)) ::
  i_ld r_ret (o_sind r_sp o_zero) ::
  i_ld r_env (o_sind r_sp o_one) :: nil.


Fixpoint tr_exp e imports valctx contctx :=
  let result_for_basic := o_hind r_arg o_zero in
  match e with

    
    
    

    | IEunit =>
        nil 

    | IEnat n =>
        i_ld r_aux (o_imm n) ::
        i_sto result_for_basic r_aux :: nil

    | IEpair x1 x2 =>
        let c1 := getVal x1 imports valctx r_ret in
        let c2 := getVal x2 imports valctx r_ret in
        c1 ++
        i_ld r_aux (o_imm 2) ::
        i_alc r_clo r_aux ::
        i_sto (o_hind r_clo o_zero) r_ret ::
        c2 ++
        i_sto (o_hind r_clo o_one) r_ret ::
        i_sto result_for_basic r_clo :: nil

    | IEfst x =>
        getVal x imports valctx r_clo ++
        i_ld r_aux (o_hind r_clo o_zero) ::
        i_sto result_for_basic r_aux :: nil

    | IEsnd x =>
        getVal x imports valctx r_clo ++
        i_ld r_aux (o_hind r_clo o_one) ::
        i_sto result_for_basic r_aux :: nil

    | IEinl x =>
        getVal x imports valctx r_ret ++
        i_ld r_aux (o_imm 2) ::
        i_alc r_clo r_aux ::
        i_ld r_aux (o_imm 0) ::
        i_sto (o_hind r_clo o_zero) r_aux ::
        i_sto (o_hind r_clo o_one) r_ret ::
        i_sto result_for_basic r_clo :: nil

    | IEinr x =>
        getVal x imports valctx r_ret ++
        i_ld r_aux (o_imm 2) ::
        i_alc r_clo r_aux ::
        i_ld r_aux (o_imm 1) ::
        i_sto (o_hind r_clo o_zero) r_aux ::
        i_sto (o_hind r_clo o_one) r_ret ::
        i_sto result_for_basic r_clo :: nil

    | IEbop op x y =>
        let c1 := getValNat x imports valctx r_ret in
        let c2 := getValNat y imports valctx r_clo in
        c1 ++ c2 ++
        i_opr op r_aux (o_reg r_ret) (o_reg r_clo) ::
        i_sto result_for_basic r_aux :: nil

    | IEfix bf bx bk e =>
        let c := tr_exp e imports (bx :: bf :: valctx) [bk] in
        i_ld r_aux (o_imm 2) ::
        i_alc r_clo r_aux ::
        i_lpc r_aux ::
        i_opr Plus r_aux (o_reg r_aux) (o_imm 7) ::
        i_sto (o_hind r_clo o_zero) r_aux ::
        i_sto (o_hind r_clo o_one) r_env ::
        i_sto result_for_basic r_clo ::
        i_opr Plus r_clo (o_reg r_aux) (o_imm (length c + 13)) ::
        i_jmp (o_reg r_clo) ::
        
        i_sto (o_sind r_sp o_zero) r_ret ::
        i_sto (o_sind r_sp o_one) r_env ::
        i_opr Plus r_sp (o_reg r_sp) (o_imm 2) ::
        i_ld r_aux (o_imm 2) ::
        i_alc r_env r_aux ::
        i_sto (o_hind r_env o_zero) r_clo ::
        i_ld r_aux (o_hind r_clo o_one) ::
        i_sto (o_hind r_env o_one) r_aux ::
        i_ld r_aux (o_imm 2) ::
        i_alc r_clo r_aux ::
        i_sto (o_hind r_clo o_zero) r_arg ::
        i_sto (o_hind r_clo o_one) r_env ::
        i_ld r_env (o_reg r_clo) ::
        c

    | IEgen bk e =>
        let c := tr_exp e imports valctx [bk] in
        i_ld r_aux (o_imm 2) ::
        i_alc r_clo r_aux ::
        i_lpc r_aux ::
        i_opr Plus r_aux (o_reg r_aux) (o_imm 7) ::
        i_sto (o_hind r_clo o_zero) r_aux ::
        i_sto (o_hind r_clo o_one) r_env ::
        i_sto result_for_basic r_clo ::
        i_opr Plus r_clo (o_reg r_aux) (o_imm (length c + 4)) ::
        i_jmp (o_reg r_clo) ::
        
        i_sto (o_sind r_sp o_zero) r_ret ::
        i_sto (o_sind r_sp o_one) r_env ::
        i_opr Plus r_sp (o_reg r_sp) (o_imm 2) ::
        i_ld r_env (o_hind r_clo o_one) ::
        c

    | IErefeq x y =>
        let c1 := getVal x imports valctx r_ret in
        let c2 := getVal y imports valctx r_clo in
        c1 ++ c2 ++
        i_opr Equal r_aux (o_reg r_ret) (o_reg r_clo) ::
        i_sto result_for_basic r_aux :: nil

    | IEcont _ _ => nil 

    
    

    | IEifnz x ea eb =>
        let cx := getVal x imports valctx r_arg in
        let ca := tr_exp ea imports valctx contctx in
        let cb := tr_exp eb imports valctx contctx in
        cx ++
        i_lpc r_clo ::
        i_opr Plus r_clo (o_reg r_clo) (o_imm (length cb + 3)) ::
        i_jnz r_arg (o_reg r_clo) ::
        cb ++ ca

    | IEcase x bx ea eb =>
        let cx := getVal x imports valctx r_arg in
        let ca := tr_exp ea imports (bx :: valctx) contctx in
        let cb := tr_exp eb imports (bx :: valctx) contctx in
        cx ++
        i_ld r_aux (o_imm 2) ::
        i_alc r_clo r_aux ::
        i_sto (o_hind r_clo o_one) r_env ::
        i_ld r_aux (o_hind r_arg o_one) ::
        i_sto (o_hind r_clo o_zero) r_aux ::
        i_ld r_env (o_reg r_clo) ::
        i_lpc r_clo ::
        i_opr Plus r_clo (o_reg r_clo) (o_imm (length ca + 4)) ::
        i_ld r_aux (o_hind r_arg o_zero) ::
        i_jnz r_aux (o_reg r_clo) ::
        ca ++ cb

    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    

    

    
    
    
    
    
    
    
    
    

    | IElet ae (inr bk) ea =>
        match ae with
          | IEcont bx eb =>
            let ca := tr_exp ea imports valctx (bk :: contctx) in
            let cb := tr_exp eb imports (bx :: valctx) contctx  in
            i_lpc r_aux ::
            i_opr Plus r_aux (o_reg r_aux) (o_imm (5 + length ca)) ::
            i_sto (o_sind r_sp o_zero) r_aux ::
            i_sto (o_sind r_sp o_one) r_env ::
            i_opr Plus r_sp (o_reg r_sp) (o_imm 2) ::
            ca ++
            
            i_ld r_clo (o_reg r_env) ::
            i_ld r_aux (o_imm 2) ::
            i_alc r_env r_aux ::
            i_sto (o_hind r_env o_zero) r_arg ::
            i_sto (o_hind r_env o_one) r_clo ::
            cb
          | _ => nil 
        end

    | IElet e1 (inl bx) e2 =>
        let c1 := tr_exp e1 imports valctx contctx in
        let c2 := tr_exp e2 imports (bx :: valctx) contctx in
        i_ld r_aux (o_imm 2) ::
        i_alc r_arg r_aux ::
        i_sto (o_hind r_arg o_one) r_env ::
        c1 ++
        i_ld r_env (o_reg r_arg) ::
        c2

    | IEapp xf x xk =>
        let c1 := getVal x imports valctx r_arg in
        let c2 := getVal xf imports valctx r_clo in
        let c3 := getCont xk contctx in
        c1 ++ c2 ++ c3 ++ i_jmp (o_hind r_clo o_zero) :: nil

    | IEinst xf xk =>
        let c1 := getVal xf imports valctx r_clo in
        let c2 := getCont xk contctx in
        c1 ++ c2 ++ i_jmp (o_hind r_clo o_zero) :: nil

    | IEret xk x =>
        let c1 := getVal x imports valctx r_arg in
        let c2 := getCont xk contctx in
        c1 ++ c2 ++ i_jmp (o_reg r_ret) :: nil

    | IEinput bx e =>
        i_inp r_aux ::
        i_ld r_arg (o_imm 2) ::
        i_alc r_arg r_arg ::
        i_sto (o_hind r_arg o_zero) r_aux ::
        i_sto (o_hind r_arg o_one) r_env ::
        i_ld r_env (o_reg r_arg) ::
        tr_exp e imports (bx :: valctx) contctx

    | IEoutput x e =>
        let c1 := getVal x imports valctx r_clo in
        let c2 := tr_exp e imports valctx contctx in
        c1 ++ i_out r_clo :: c2

    | IEnew x bx e =>
        let c1 := getVal x imports valctx r_arg in
        let c2 := tr_exp e imports (bx :: valctx) contctx in
        c1 ++
        i_ld r_aux (o_imm 1) ::
        i_alc r_clo r_aux ::
        i_sto (o_hind r_clo o_zero) r_arg ::
        i_ld r_aux (o_imm 2) ::
        i_alc r_arg r_aux ::
        i_sto (o_hind r_arg o_one) r_env ::
        i_sto (o_hind r_arg o_zero) r_clo ::
        i_ld r_env (o_reg r_arg) ::
        c2

    | IEread x bx e =>
        let c1 := getVal x imports valctx r_arg in
        let c2 := tr_exp e imports (bx :: valctx) contctx in
        c1 ++
        i_ld r_aux (o_imm 2) ::
        i_alc r_clo r_aux ::
        i_sto (o_hind r_clo o_one) r_env ::
        i_ld r_aux (o_hind r_arg o_zero) ::
        i_sto (o_hind r_clo o_zero) r_aux ::
        i_ld r_env (o_reg r_clo) ::
        c2

    | IEwrite x y e =>
        let c1 := getVal x imports valctx r_clo in
        let c2 := getVal y imports valctx r_arg in
        let c3 := tr_exp e imports valctx contctx in
        c1 ++ c2 ++
        i_sto (o_hind r_clo o_zero) r_arg ::
        c3
  end.

Fixpoint tr_toplvl v imports :=
  match v with
    | IEfix bf bx bk e =>
        i_sto (o_sind r_sp o_zero) r_ret ::
        i_sto (o_sind r_sp o_one) r_env ::
        i_opr Plus r_sp (o_reg r_sp) (o_imm 2) ::
        i_ld r_aux (o_imm 2) ::
        i_alc r_aux r_aux :: 
        i_sto (o_hind r_aux o_zero) r_clo ::
        i_ld r_env (o_imm 2) ::
        i_alc r_env r_env :: 
        i_sto (o_hind r_env o_zero) r_arg ::
        i_sto (o_hind r_env o_one) r_aux ::
        tr_exp e imports [bx; bf] [bk]
    | IEgen bk e =>
        i_sto (o_sind r_sp o_zero) r_ret ::
        i_sto (o_sind r_sp o_one) r_env ::
        i_opr Plus r_sp (o_reg r_sp) (o_imm 2) ::
        tr_exp e imports nil [bk]
    | _ => nil
  end.

Fixpoint tr_mod F (M: lang_mid.module) : lang_tgt.module :=
  fun laddr imports =>
    match M with
      | nil => nil
      | (f, v) :: M' =>
          let seg := (laddr, map enc (tr_toplvl v (combine F imports))) in
          let lsegs := tr_mod (F ++ [f]) M' (laddr + 1 + length (snd seg)) (imports ++ [laddr]) in
          (f, seg) :: lsegs
    end.
