(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List.
Require Import vgtac basis myclassical.
Require Import lang_common lang_spec lang_mid lang_tgt.
Require Import model_common model_untyped model_properties_untyped.
Require Import gw_common gw_untyped.
Require Import codegen codegen_correct_aux codegen_correct_atom codegen_correct_ctrl.
Require Import ssr util.

Set Implicit Arguments. 
Global Set Bullet Behavior "Strict Subproofs".




(** The translation preserves well-formedness. *)

Theorem tr_mod_wf: forall FI FO m
    (WF: mid_wfmod FI m FO),
  tgt_wfmod (tr_mod FI m) FO.
Proof.
  i; ginduction WF; [by apply tgt_wfmod_nil|]; rr; splits; rrdes IHWF.
  - i; s in INa; s in INb; des; clarify; eauto.
    + apply pmvalid_plus_plus.
      * s; apply pmvalid_cell_plus; first (by auto); i.
        apply pmcells_readD in RD; des; omega.
      * destruct sega; apply BND in INa; des; s.
        apply pmvalid_plus_plus.
        { apply pmvalid_cell_plus; first (by auto); i.
          apply pmread_cellD in RD; des; omega.
        }
        { apply pmvalid_cell_plus; first (by auto); i.
          apply pmcells_readD in RD; des; omega.
        }
        { apply pmvalid_cell_plus; first (by auto); i.
          apply pmcells_readD in RD; des; omega.
        }
      * s. destruct sega; apply BND in INa; des; s.
        apply pmvalid_plus_plus.
        { rewrite pmplus_comm.
          apply pmvalid_cell_plus; first (by auto); i.
          apply pmcells_readD in RD; des; omega.
        }
        { rewrite pmplus_comm; apply pmcells_plus_cells; omega. }
        { apply pmvalid_cell_plus; first (by auto); i.
          apply pmcells_readD in RD; des; omega.
        }
    + 
      apply pmvalid_plus_plus.
      * s; apply pmvalid_cell_plus; first (by auto); i.
        apply pmcells_readD in RD; des; omega.
      * destruct segb; apply BND in INb; des; s.
        apply pmvalid_plus_plus.
        { apply pmvalid_cell_plus; first (by auto); i.
          apply pmread_cellD in RD; des; omega.
        }
        { apply pmvalid_cell_plus; first (by auto); i.
          apply pmcells_readD in RD; des; omega.
        }
        { apply pmvalid_cell_plus; first (by auto); i.
          apply pmcells_readD in RD; des; omega.
        }
      * s. destruct segb; apply BND in INb; des; s.
        apply pmvalid_plus_plus.
        { rewrite pmplus_comm.
          apply pmvalid_cell_plus; first (by auto); i.
          apply pmcells_readD in RD; des; omega.
        }
        { apply pmcells_plus_cells; omega. }
        { apply pmvalid_cell_plus; first (by auto); i.
          apply pmcells_readD in RD; des; omega.
        }
  - s; i; apply in_consD in IN; des.
    + clarify; split; [by eauto|].
      unfold msize; s; omega.
    + apply BND in IN; des.
      split; [omega|]; eapply le_trans; [by eauto|].
      unfold msize; s; omega.
  - i; destruct na, nb; s in INa; s in INb; [by eauto| | |by eauto].
    + exfalso; unfold value in INa; clarify.
      apply nth_error_someD in INb; apply (in_map fst) in INb.
      rewrite -OUT in INb; ss; eapply mid_wfmod_uniq; last first; eauto.
      eapply in_or_app; s; eauto.
    + exfalso; unfold value in INb; clarify.
      apply nth_error_someD in INa; apply (in_map fst) in INa.
      rewrite -OUT in INa; ss; eapply mid_wfmod_uniq; last first; eauto.
      eapply in_or_app; s; eauto.
  - i; s; f_equal; eauto.
Qed.



Lemma tr_exp_correct: forall e G
    (WF: wfexp G e),
  tm_psim G (fun imports => map enc (tr_exp e imports (tvars G) (kvars G)))
            e.
Proof.
  intros ?; remember (size e) as n; generalize dependent e.
  induction n using (well_founded_ind lt_wf).
  destruct e; vauto; i; ss; des; subst
  ; try (hexploit (H (size e)); [M|M|M|]; (Mdo by eauto); i)
  ; try (hexploit (H (size e1)); [omega|M|M|]; (Mdo by eauto); i)
  ; try (hexploit (H (size e2)); [omega|M|M|]; (Mdo by eauto); i).
  - apply tr_exp_correct_new; eauto.
  - apply tr_exp_correct_read; eauto.
  - apply tr_exp_correct_write; eauto.
  - apply tr_exp_correct_input; eauto.
  - apply tr_exp_correct_output; eauto.
  - apply tr_exp_correct_ifnz; eauto.
  - apply tr_exp_correct_case; eauto.
  - apply tr_exp_correct_ret; eauto.
  - apply tr_exp_correct_app; eauto.
  - apply tr_exp_correct_inst; eauto.
  - destruct bx; des.
    + hexploit (H (size e2)); [s; omega|M|M|]; (Mdo by eauto); i.
      destruct e1
      ; try (by hexploit tr_exp_correct_atom
             ; [by eauto|by ii; vauto|i; eapply H; last first; eauto; omega|]
             ; i; apply tr_exp_correct_let_termvar; eauto).
      ii; econs; i; tm_prepare; pfold; econs; i; destruct cfb.
      rrdes XTRb; clarify; invc CNF; des; subst; noemp; rrdes MB.
      s in Ec; clarify.
      eapply esim_err.
      * match goal with |- context[(_, ?ee)] => instantiate (1:=(None, (envb, ee))) end.
        econs 1; eapply IRerr; vauto; i; invc FH.
      * eauto.
    + destruct e1;
      try by ii; econs; i; tm_prepare; pfold; econs; i; destruct cfb
           ; rrdes XTRb; clarify; invc CNF; des; subst; noemp; rrdes MB
           ; s in Ec; clarify
           ; econstructor (try match goal with |- context[(_, ?ee)] => instantiate (1:=(None, (envb, ee))) end
                           ; eauto using esim_err; econs 1; eapply IRerr; vauto; i; invc FH).
      
      hexploit (H (size e1)); [s; omega|M|M|]; (Mdo by eauto); i.
      hexploit (H (size e2)); [s; omega|M|M|]; (Mdo by eauto); i.
      eapply tr_exp_correct_let_contvar; eauto.
Grab Existential Variables.
exact (wfo_nat 0). exact (wfo_nat 0). exact (wfo_nat 0). exact (wfo_nat 0).
exact (wfo_nat 0). exact (wfo_nat 0). exact (wfo_nat 0). exact (wfo_nat 0).
exact (wfo_nat 0). exact (wfo_nat 0). exact (wfo_nat 0). exact (wfo_nat 0).
exact (wfo_nat 0). exact (wfo_nat 0). exact (wfo_nat 0). exact (wfo_nat 0).
exact (wfo_nat 0). exact (wfo_nat 0). exact (wfo_nat 0). exact (wfo_nat 0).
exact (wfo_nat 0). exact (wfo_nat 0). exact (wfo_nat 0).
Qed.



Lemma tr_toplvl_correct: forall FI e ancha suba subb s codeframe
    (SUBa: map fst suba = FI)
    (SUBb: map fst subb = FI)
    (GT: is_good_thunk e)
    (WF: wfatom (map (fun l : label => inl (inl l)) FI) e)
    (FUN: tm_db s (ts_fun (S ancha)) ancha : Prop)
    (CODE: tm_lhp s = {S ancha |=> map enc (tr_toplvl e (combine FI (map snd suba)))} ## codeframe)
    (VLD: pmvalid (tm_lhp s)),
  @tlsim _ _ tm_gw mylw (map fst suba) suba subb s
        ancha
        (IVclo (map_pair (inl<*>inl) id subb) e).
Proof.
  ii; destruct e; vauto; s in WF.
  - hexploit tr_exp_correct; eauto; i. s in X.
    rewrite kvars_purge in X; specialize (X wfo_zero); rrdes X; rename X into IH.
    apply vclf_fix; s; eauto; first (by econs; eapply LE; eauto).
    evar (i0: wfo_set (wlift mylw)).             
    eapply gk_lsim; eauto.
    econs.
    { econs.
      - econs; rr; eauto.
      - s. econs. i. apply GT. inv FH.
    }
    instantiate (1:=i0).
    pcofix CIH. unfold i0 in *. clear i0.
    i; invc STK; [|by invc INSTb RF CB VR].
    eapply pesim_gknow; i; eapply pesim_gknow_; i.
    invc APPa APPb.
    invc RF CB VR. rrdes CA.
    destruct s' as [[[[]]]]; clean.
    ren PC.
    destruct s3.
    tm_prepare.

    hexploit (tgt_stack CPa 2); i; des; s in FST; subst freest.
    des; repeat (destruct vs; [by vauto|]); destruct vs; [|by vauto].
    rewrite <-!pmplus_assoc in *; clear NST.

    hexploit (ext_trans _ _ _ EXT _ LE); i.
    dup X; rrdes X X.

    hexploit (@pmplus_from_read _ _ _ (snd (snd s)) h); [by eauto|by eauto|].
    i; des; subst h.

    
    clear v' CA; rrdes X X; apply X in FUN.
    hexploit tgt_vquery_fun_uniq; [by eauto|exact FUN|exact PC0|]; i; subst.

    eapply pesim_dance; first (by eauto); clear RLa0.
    + clear SV HV FIN.

      dance_init.
      { eapply Sok; [by vauto|eapply Ssto].
        - unfold mrf; eapply pmread_plus; eauto.
          rewrite CODE; s; rewrite !pmcells_cons; do 2 pmplus_eq''.
        - rewrite pmcells_cons -pmplus_assoc.
          replace (mrf r_sp) with (add_offset (mrf r_sp) o_zero)
          ; last (by s; tmext_simpl; eauto); eauto.
      }
      { s; pmcells_fold.
        depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
      }

      dance_init.
      { eapply Sok; [by vauto|eapply Ssto].
        - eapply pmread_plus; eauto.
          rewrite CODE; s; rewrite !pmcells_cons; pmplus_eq''.
        - rewrite 2!pmcells_cons -2!pmplus_assoc pmplus_swap.
          replace (S (mrf r_sp)) with (add_offset (mrf r_sp) o_one)
          ; last (by s; tmext_simpl; eauto); eauto.
      }
      { s; pmcells_order; pmcells_fold.
        depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
      }

      dance_init.
      { eapply Sok; [by vauto|eapply Sopr].
        - eapply pmread_plus; eauto.
          rewrite CODE; s; rewrite !pmcells_cons; pmplus_eq'.
        - eauto.
        - eauto.
        - eauto.
      }
      { instantiate (1 := depmach0).
        eapply tgt_mcplus; first by s; eauto.
        rr; esplits; s; try pmplus_eq; eauto.
      }

      dance_init.
      { eapply Sok; [by vauto|eapply Sld].
        - eapply pmread_plus; eauto.
          rewrite CODE; s; rewrite !pmcells_cons; pmplus_eq''.
        - eauto.
      }
      { depmcplus0. }

      dance_init.
      { eapply Sok; [by vauto|eapply Salc].
        - eapply pmread_plus; eauto.
          rewrite CODE; s; rewrite !pmcells_cons; pmplus_eq''.
        - eauto.
      }
      { depmcplus1; eauto using alloc_valid. }
      s; setnewloc.

      dance_init.
      { eapply Sok; [by vauto|eapply Ssto].
        - eapply pmread_plus; eauto.
          rewrite CODE; s; rewrite !pmcells_cons; pmplus_eq''.
        - tmext_simpl; rewrite (pmcells_cons newloc) -pmplus_assoc.
          replace newloc with (add_offset (mrf r_aux) o_zero)
          ; last (by unfold mrf; s; tmext_simpl; eauto); eauto.
      }
      { s; tmext_simpl; pmcells_fold.
        depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
      }
      s; fold newloc.

      dance_init.
      { eapply Sok; [by vauto|eapply Sld].
        - eapply pmread_plus; eauto.
          rewrite CODE; s; rewrite !pmcells_cons; pmplus_eq''.
        - eauto.
      }
      { depmcplus0. }
      
      dance_init.
      { eapply Sok; [by vauto|eapply Salc].
        - eapply pmread_plus; eauto.
          rewrite CODE; s; rewrite !pmcells_cons; pmplus_eq''.
        - eauto.
      }
      { depmcplus1; eauto using alloc_valid. }
      s; setnewloc.

      dance_init.
      { eapply Sok; [by vauto|eapply Ssto].
        - eapply pmread_plus; eauto.
          rewrite CODE; s; rewrite !pmcells_cons; pmplus_eq''.
        - rewrite (pmcells_cons newloc0) -pmplus_assoc.
          replace newloc0 with (add_offset (mrf r_env) o_zero)
          ; last (by unfold mrf; s; tmext_simpl; eauto); eauto.
      }
      { s; tmext_simpl; pmcells_fold.
        depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
      }
      s; fold newloc0.

      dance_init.
      { eapply Sok; [by vauto|eapply Ssto].
        - eapply pmread_plus; eauto.
          rewrite CODE; s; rewrite !pmcells_cons; pmplus_eq''.
        - rewrite (pmcells_cons newloc0) -pmcells_sng -pmplus_assoc (pmplus_swap {newloc0|->_}).
          replace (S newloc0) with (add_offset (mrf r_env) o_one)
          ; last (by unfold mrf; s; tmext_simpl; eauto); eauto.
      }
      { s; tmext_simpl; pmcells_order; pmcells_fold.
        depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
      }

      eapply dance_O.
      instantiate (1 := fun _ _ _ _ cf _ _ cf' _ _ => depmach2 cf cf'); s; eauto.

    + i; rrdes XTRb RLb0; clarify.
      esplits; first 1 last.
      { right; split; eauto using wfo_le_n_add. }
      { s in Ec. clarify. inst_pairs.  s. noemp. rr; esplits; eauto. }
    
      i; frames_dep 4 RLa XTRa cfas.
      frames_dep 2 RLa XTRa cfas.
      frames_nodep RLa cfas.

      destruct ea'; clear SV HV FIN.
      destruct_tgt_real RLa.
      simpl lang_tgt.chp in *.  
      setnewloc.
      setcrf mrf'.
      set (s' := (mrf', P, s1, s2, ({mrf r_sp |=> [mrf r_ret; mrf r_env]} ## s3,
                  {newloc |=> [mrf r_arg; newloc0]} ##
                  {newloc0 |=> [mrf r_clo; 0]} ##
                  snd (snd s) ## h''))).

      assert (VLD': pmvalid ({newloc |=> [mrf r_arg; newloc0]} ##
                             {newloc0 |=> [mrf r_clo; 0]} ## snd (snd s) ## h'')).
      { clear -HV.
        rewrite -!(pmplus_swap hp) -!(pmplus_swap refha) in HV.
        eapply pmvalid_plusD, pmvalid_plusD; eauto.
      }

      assert (EXT': ext (wlift mylw) s' (mrf, P, s1, s2, (s3, snd (snd s) ## h''))).
      { clear -VLD'.
        s; splits; tmext_simpl; auto.
        i; eapply pmread_plus; eauto.
        set (tmp := snd (snd s) ## h''); pmplus_eq'.
      }

      s in Ec; clarify.

      exists s'; esplits; [exact EXT'|].
      rewrite tgt_csplit mid_csplit.
      eapply pesim_release; last first; try by rr; s; eauto.
      { unfold s'.
        rr; esplits.
        - rr; esplits; eauto.
          + s; rr; esplits; eauto.
            * clean.
              instantiate (1 := {| cpc := Nothing; crf := Just mrf'; cst := freest0; chp := hp |}).
              rrdes CPa; s in HP.
              unfold mrf'; rr; splits; s; tmext_simpl; auto.
              s in ST.
              assert (XX := tgt_stack_transfer _ _ _ ST). s in XX. eauto.
          + s; eapply tm_refconfrel_mon
            ; [eapply tm_refconfrel'_smon; s; eauto|]; inst
            ; s; i; eapply gk_mon; eauto; rrdes EXT'; rr; eauto.
      - rr; esplits; eauto.
      - s; noemp; f_equal.
        + rewrite pmplus_swap; eauto.
        + rewrite (pmplus_assoc hp); set (tmp := hp ## _).
          rewrite -!(pmplus_swap tmp); eauto.
      - noemp; auto.
      }


      eapply paco9_mon; [eapply IH|by eauto]; eauto.
      * i; s in IN; des; vauto.
        apply purge_inD in IN.
        apply in_map_iff in IN; des. clarify.
        hexploit list_assoc_map; eauto; i; des.
        dup IN0; rewrite -SUBb in IN0.
        hexploit list_assoc_map; eauto; i; des.
        hexploit RS; eauto; i; des.
        esplits; eauto.
        { rewrite list_assoc_cons_neq; [|by ii; vauto].
          rewrite list_assoc_cons_neq; [|by ii; vauto].
          rewrite list_assoc_cons_neq; [|by ii; vauto].
          eapply list_assoc_map_pair; eauto.
          i; clarify.        
        }
        eapply vclos_mon; [econs; eapply vclgen_thunk; eauto| |]; eauto using ext_trans.
        i. eapply LE0; eapply gk_mon; eauto using ext_trans.
      * instantiate (1 := (bx0, mrf r_arg) :: (bf0, mrf r_clo) :: nil).
        i; apply in_consD in IN; des; vauto.
        apply in_consD in IN; des.
        { clarify.
          rewrite list_assoc_cons_eq.
          rewrite list_assoc_cons_neq; [|by vauto].
          rewrite list_assoc_cons_eq.
          esplits; eauto.
          eapply vclos_mon; eauto using ext_trans.
          i; eapply gk_mon; eauto using ext_trans.
        }
        rewrite list_assoc_cons_neq; [|by ii; vauto].
        apply in_consD in IN; des.
        { clarify.
          rewrite list_assoc_cons_eq.
          rewrite list_assoc_cons_neq; [|by ii; vauto].
          rewrite list_assoc_cons_neq; [|by ii; vauto].
          rewrite list_assoc_cons_eq.
          esplits; eauto.
          econs.
          apply vclgen_thunk, vclf_fix.
          { unfold s'; econs; rr; eauto. }
          { rr; eauto. }

          eapply gk_lsim; eauto.
          econs.
          { econs.
            - unfold s'; econs; rr; eauto.
            - s. econs. i. apply GT. inv FH.
          }
          
          i; eapply gres_incl_id. right.
          eapply CIH; eauto using ext_trans.
        }
        exfalso; apply purge_inD, in_map_iff in IN; des; vauto.
      * unfold s', mrf'; s; tmext_simpl; esplits; auto
        ; try (by eapply pmread_plus; eauto; rewrite !pmcells_cons; pmplus_eq').
      * s; rewrite tvars_purge. 
        clear; induction (map fst suba); auto.
      * unfold s'; s; eauto.
      * s; rewrite kvars_purge; s; eauto.
      * unfold s', mrf'; s; tmext_simpl.
        rewrite kvars_purge; s; omega.
      * unfold s', mrf'; tmext_simpl; s.
        econs.
        { s; eauto. }
        { rewrite kvars_purge; s.
          econs. s. tmext_simpl. omega.
        }
        { s; tmext_simpl.
          rewrite plus_comm plus_minus_eq; last omega; s.
          decomp_pmvalid SV.
          eapply pmread_plus; eauto.
          rewrite pmcells_cons; pmplus_eq'.
        }
        { s; tmext_simpl.
          rewrite plus_comm; s; rewrite plus_comm plus_minus_eq; last omega; s.
          decomp_pmvalid SV.
          eapply pmread_plus; eauto.
          rewrite 2!pmcells_cons; pmplus_eq'.
        }
        { rewrite list_assoc_cons_eq; eauto. }

        i. invc RETa.
        invc RETb.
        invc RK.                        
        eapply tm_pesim_return; eauto.
        {
          rrdes EXT0 EXT0 EXT0 EXT0; split; [repeat split|]; s; eauto.
          - ss; rewrite SP; tmext_simpl.
            rewrite plus_comm plus_minus_eq; s; auto.
          - split.
            + ii; s in EXT1; unnw; hexploit EXT1; eauto.
              eapply pmread_plus; eauto.
              set (tmp := _ ## h''); pmplus_eq'.
            + ss; rewrite SP in LST; tmext_simplH LST.
              rewrite plus_comm plus_minus_eq in LST; last auto; s in LST.
              rewrite (pmplus_comm (fst _)) in LST; decomp_pmvalid SV.
              apply pmcells_uniq in LST; des; eauto.
        }
      * unfold s'; s; econs; symmetry.
        match goal with |- context[tvars ?x] => replace (tvars x) with (@nil termvar) end.
        rewrite -(combine_fst_snd suba).
        rewrite CODE; s; rewrite !pmcells_cons.
        pmplus_eq'.
        clear; induction (map fst suba); auto.

  - hexploit tr_exp_correct; eauto; i. s in X.
    rewrite kvars_purge in X; specialize (X wfo_zero); rrdes X; rename X into IH.
    apply vclf_gen; s; eauto; first (by econs; eapply LE; eauto).
    evar (i0: wfo_set (wlift mylw)).             
    eapply gk_lsim; eauto.
    econs.
    { econs 2.
      - econs; rr; eauto.
      - s. econs. i. apply GT. inv FH.
    }
    instantiate (1:=i0).
    pcofix CIH. unfold i0 in *. clear i0.
    i; invc STK; [by invc APPb RF CB VR|].
    eapply pesim_gknow; i; eapply pesim_gknow_; i.
    invc INSTa INSTb.
    invc RF CB VR. rrdes CA.
    destruct s' as [[[[]]]]; clean.
    ren PC.
    destruct s3.
    tm_prepare.

    hexploit (tgt_stack CPa 2); i; des; s in FST; subst freest.
    des; repeat (destruct vs; [by vauto|]); destruct vs; [|by vauto].
    rewrite <-!pmplus_assoc in *; clear NST.

    hexploit (ext_trans _ _ _ EXT _ LE); i.
    dup X; rrdes X X.

    hexploit (@pmplus_from_read _ _ _ (snd (snd s)) h); [by eauto|by eauto|].
    i; des; subst h.

    
    clear v' CA; rrdes X X; apply X in FUN.
    hexploit tgt_vquery_fun_uniq; [by eauto|exact FUN|exact PC0|]; i; subst.

    eapply pesim_dance; first (by eauto); clear RLa0.
    + clear SV HV FIN.

      dance_init.
      { eapply Sok; [by vauto|eapply Ssto].
        - unfold mrf; eapply pmread_plus; eauto.
          rewrite CODE; s; rewrite !pmcells_cons; do 2 pmplus_eq''.
        - rewrite pmcells_cons -pmplus_assoc.
          replace (mrf r_sp) with (add_offset (mrf r_sp) o_zero)
          ; last (by s; tmext_simpl; eauto); eauto.
      }
      { s; pmcells_fold.
        depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
      }

      dance_init.
      { eapply Sok; [by vauto|eapply Ssto].
        - eapply pmread_plus; eauto.
          rewrite CODE; s; rewrite !pmcells_cons; pmplus_eq''.
        - rewrite 2!pmcells_cons -2!pmplus_assoc pmplus_swap.
          replace (S (mrf r_sp)) with (add_offset (mrf r_sp) o_one)
          ; last (by s; tmext_simpl; eauto); eauto.
      }
      { s; pmcells_order; pmcells_fold.
        depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
      }

      dance_init.
      { eapply Sok; [by vauto|eapply Sopr].
        - eapply pmread_plus; eauto.
          rewrite CODE; s; rewrite !pmcells_cons; pmplus_eq'.
        - eauto.
        - eauto.
        - eauto.
      }
      { instantiate (1 := depmach0).
        eapply tgt_mcplus; first by s; eauto.
        rr; esplits; s; try pmplus_eq; eauto.
      }

      eapply dance_O; instantiate (1 := fun _ _ _ => _).
      s; eauto.

    + i; rrdes XTRb RLb0; clarify.
      esplits; first 1 last.
      { right; split; eauto using wfo_le_n_add. }
      { s in Ec. clarify. inst_pairs.  s. noemp. rr; esplits; eauto. }
    
      i; frames_nodep RLa cfas.

      destruct ea'; clear SV HV FIN.
      destruct_tgt_real RLa.

      setcrf mrf'.
      set (s' := (mrf', P, s1, s2, ({mrf r_sp |=> [mrf r_ret; mrf r_env]} ## s3, snd (snd s) ## h''))).

      assert (EXT': ext (wlift mylw) s' (mrf, P, s1, s2, (s3, snd (snd s) ## h''))).
      { s; splits; tmext_simpl; auto. }

      s in Ec. clarify.

      exists s'; esplits; [exact EXT'|].
      rewrite tgt_csplit mid_csplit.
      eapply pesim_release; last first; try by rr; s; eauto.
      { unfold s'.
        rr; esplits.
        - rr; esplits; eauto.
          + s; rr; esplits; eauto.
            * clean.
              instantiate (1 := {| cpc := Nothing; crf := Just mrf'; cst := freest0; chp := hp |}).
              rrdes CPa; s in HP.
              unfold mrf'; rr; splits; s; tmext_simpl; auto.
              s in ST.
              assert (XX := tgt_stack_transfer _ _ _ ST). s in XX. eauto.
          + s; eapply tm_refconfrel_mon
            ; [eapply tm_refconfrel'_smon; s; eauto|]; inst
            ; s; i; eapply gk_mon; eauto; rrdes EXT'; rr; eauto.
      - rr; esplits; eauto.
      - s; noemp; f_equal.
        + rewrite pmplus_swap; eauto.
        + rewrite (pmplus_assoc hp); set (tmp := hp ## _). auto.
      - noemp; auto.
      }

      eapply paco9_mon; [eapply IH|by eauto]; eauto.
      * i; s in IN; des; vauto.
        apply purge_inD in IN.
        apply in_map_iff in IN; des. clarify.
        hexploit list_assoc_map; eauto; i; des.
        dup IN0; rewrite -SUBb in IN0.
        hexploit list_assoc_map; eauto; i; des.
        hexploit RS; eauto; i; des.
        esplits; eauto.
        { rewrite list_assoc_cons_neq; [|by ii; vauto].
          eapply list_assoc_map_pair; eauto.
          i; clarify.        
        }
        eapply vclos_mon; [econs; eapply vclgen_thunk; eauto| |]; eauto using ext_trans.
        i. eapply LE0; eapply gk_mon; eauto using ext_trans.
      * instantiate (1 := nil).
        i; apply in_consD in IN; des; vauto.
        exfalso; apply purge_inD, in_map_iff in IN; des; vauto.
      * unfold s', mrf'; s; tmext_simpl; esplits; auto
        ; try (by eapply pmread_plus; eauto; rewrite !pmcells_cons; pmplus_eq').
      * s; rewrite tvars_purge. 
        clear; induction (map fst suba); auto.
      * unfold s'; s; eauto.
      * s; rewrite kvars_purge; s; eauto.
      * unfold s', mrf'; s; tmext_simpl.
        rewrite kvars_purge; s; omega.
      * unfold s', mrf'; tmext_simpl; s.
        econs.
        { s; eauto. }
        { rewrite kvars_purge; s.
          econs. s. tmext_simpl. omega.
        }
        { s; tmext_simpl.
          rewrite plus_comm plus_minus_eq; last omega; s.
          decomp_pmvalid SV.
          eapply pmread_plus; eauto.
          rewrite pmcells_cons; pmplus_eq'.
        }
        { s; tmext_simpl.
          rewrite plus_comm; s; rewrite plus_comm plus_minus_eq; last omega; s.
          decomp_pmvalid SV.
          eapply pmread_plus; eauto.
          rewrite 2!pmcells_cons; pmplus_eq'.
        }
        { rewrite list_assoc_cons_eq; eauto. }

        i. invc RETa.
        invc RETb.
        invc RK.                        
        eapply tm_pesim_return; eauto.
        {
          rrdes EXT0 EXT0 EXT0 EXT0; split; [repeat split|]; s; eauto.
          - ss; rewrite SP; tmext_simpl.
            rewrite plus_comm plus_minus_eq; s; auto.
          - split.
            + ii; s in EXT1; unnw; hexploit EXT1; eauto.
            + ss; rewrite SP in LST; tmext_simplH LST.
              rewrite plus_comm plus_minus_eq in LST; last auto; s in LST.
              rewrite (pmplus_comm (fst _)) in LST; decomp_pmvalid SV.
              apply pmcells_uniq in LST; des; eauto.
        }
      * unfold s'; s; econs; symmetry.
        match goal with |- context[tvars ?x] => replace (tvars x) with (@nil termvar) end.
        rewrite -(combine_fst_snd suba).
        rewrite CODE; s; rewrite !pmcells_cons.
        pmplus_eq'.
        clear; induction (map fst suba); auto.
Qed.


Theorem tr_mod_correct: forall FI FO m
    (WF: mid_wfmod FI m FO),
  msim tm_gw FI FO (tr_mod FI m) m.
Proof.
  exists mylw; i.

  destruct ciga, cila; rrdes LMa.
  simpl lang_tgt.cpc in *; simpl lang_tgt.crf in *.
  simpl lang_tgt.cst in *; simpl lang_tgt.chp in *.
  clarify.

  destruct cigb, cilb; rrdes LMb; clarify.

  set (db := fun q v => exists v', q = ts_fun v' /\
         pmread (hp_imports ## loadgh (map snd (tr_mod (map fst suba) m ancha (map snd suba)))) v v').
  set (sig := (rf, db, tt, refst_emp) : state tm_gw).
  set (codeframe := pmemp : heap).
  set (sil := (pmemp, codeframe ## loadlh (map snd (tr_mod (map fst suba) m ancha (map snd suba)))) : state mylw).

  exists (sig, sil); splits.
  - apply mylw_stable.
  - i; rr; econs; exists cemp; econs; exists cemp; splits; noemp; last first; eauto.
    + unfold sig; s; rr; esplits; s; unfold mid_cemp, tgt_cemp; eauto
      ; i; try apply pmread_empD in ALC; vauto.
    + rr; splits; econs; s; rr; splits; auto.
      * rewrite GST; split; [by eauto|i; omega].
      * i; des; rrdes IN; subst; s; eauto.
  - rr; esplits; s; unfold codeframe; noemp; eauto.
  - i; unfold sig in *; rrdes FH; vauto.
  - assert (VLD: pmvalid (codeframe ## loadlh (map snd (tr_mod (map fst suba) m ancha (map snd suba))))).
    { unfold codeframe; noemp; split_pmvalid LDJ; auto. }

    clear dependent cst; generalize dependent codeframe.
    remember (map fst suba) as FI.
    i; ginduction WF; i; vauto.

    hexploit tr_mod_wf; [by econs 2; eauto|]; intro WFM.
    hexploit (@loadgh_loadlh_valid (tr_mod FI ((f, e) :: m)) (f :: FO) WFM ancha (map snd suba)).
    intro VLDM. s in WFM.

    assert (SIM: @tlsim _ _ tm_gw mylw (map fst suba) suba subb (sig, sil)
                       ancha
                       (IVclo (map (fun p : label * val => (inl (inl (fst p)), snd p)) subb) e)).
    { eapply (@tr_toplvl_correct FI e ancha suba subb); eauto.
      - unfold sig, db; s; esplits; eauto.
        eapply pmread_plus; eauto; pmplus_eq'.
      - unfold sil; symmetry; s; pmplus_eq'.
    }

    apply in_consD in IN; des.
    + 
      subst f0; esplits.
      * rr; rewrite list_assoc_cons_eq; eauto.
      * rr; s; des_if; vauto.
      * rewrite HeqFI; eauto.
    + 
      simpl (vload mid_lsi); unfold mid_vload; simpl mid_vload'.
      des_if; [congruence|].
      match goal with |- context[mid_vload' _ ?sub _] =>
        set (subb' := sub)
      end.
      
      hexploit IHWF.
      * eauto.
      * instantiate (1 := hp_imports ## {ancha |-> S ancha}).
        instantiate (1 := suba ++ [(f, ancha)]).
        split; i.
        { s in GDJ; decomp_pmvalid GDJ.
          rewrite map_app in H; apply in_app_or in H; des.
          - apply GDOM in H; des; econs; eapply pmread_plus; eauto.
          - s in H; des; vauto; rewrite -H.
            econs; eapply pmread_plus; eauto; try pmplus_eq'.
        }
        { des; apply pmread_plusD in H; des.
          - rewrite map_app; apply in_or_app; left.
            apply GDOM; eauto.
          - rewrite map_app; apply in_or_app; right.
            apply pmread_cellD in H; des; s; eauto.
        }
      * rewrite HeqFI map_app; auto.
      * instantiate (1 := subb'); unfold subb'.
        rewrite map_app SUBb; auto.
      * rewrite -pmplus_assoc map_app; s; s in GDJ; s in LDJ.
        decomp_pmvalid GDJ; decomp_pmvalid LDJ; apply pmvalid_plus_plus; eauto.
        s in VLDM; rewrite -pmplus_assoc !(pmplus_swap {ancha |-> _}) in VLDM.
        apply pmvalid_plusD in VLDM; des; apply pmvalid_plusD in VLDM0; des; auto.
      * rewrite -pmplus_assoc map_app; s; s in GDJ; s in LDJ.
        decomp_pmvalid GDJ; decomp_pmvalid LDJ; apply pmvalid_plus_plus; eauto.
      * instantiate (1 := codeframe ## {S ancha |=> map enc (tr_toplvl e (combine FI (map snd suba)))}).
        rewrite -pmplus_assoc map_app; s; eauto.
      * s in IN; des; vauto.
      * i; des; esplits.
        { rr; simpl tr_mod.
          rewrite list_assoc_cons_neq; [|by vauto].
          rrdes H; rewrite map_app in H; eauto.
        }
        { rrdes H0; eauto. }
        ii; eapply H1; eauto.
        { unfold sig, db, sil in LE.
          simpl tr_mod in LE; simpl loadgh in LE; simpl loadlh in LE.
          rewrite -!pmplus_assoc map_app; auto.
        }
        i; apply list_assoc_app_or in INsuba.
        apply list_assoc_app_or in INsubb; des.
        { apply in_app_xor in INienv; des; [by eauto|].
          exfalso; apply INienv; apply list_assoc_someD in INsuba.
          eapply (in_map fst) in INsuba; rewrite HeqFI; auto.
        }
        { apply list_assoc_someD in INsubb.
          apply list_assoc_noneD in INsuba; vauto.
          rewrite -HeqFI -SUBb; eapply (in_map fst) in INsubb; auto.
        }
        { apply list_assoc_someD in INsuba.
          apply list_assoc_noneD in INsubb; vauto.
          rewrite SUBb HeqFI; eapply (in_map fst) in INsuba; auto.
        }
        { apply list_assoc_someD in INsuba0; apply list_assoc_someD in INsubb0.
          s in INsuba0; s in INsubb0; des; vauto; subst f1 va' vb'.
          rewrite HeqFI in RS.
          eapply vclos_thunk_mon; first (eapply SIM); eauto using (gk_lsim GK).
        }
Grab Existential Variables.
  exact tt.
Qed.






