(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List Omega Max.
Require Import vgtac basis countable myclassical paco9.
Require Import types lang_spec lang_common lang_src_lsi lang_tgt.
Require Import model_common model model_properties.
Require Import gw gw_common hcomp_esim.

Set Implicit Arguments.
Global Set Bullet Behavior "Strict Subproofs".

Local Notation cid := constructive_indefinite_description.


Section msim.

  Variable mat mbt: lang_tgt.module.
  Variable mas mbs: lang_src.module.
  Variable ienv oenva oenvb: list (label * flextyp).
  Variable WFat: tgt_wfmod mat (map fst oenva).
  Variable WFas: src_wfmod mas (map fst oenva).
  Variable WFbt: tgt_wfmod mbt (map fst oenvb).
  Variable WFbs: src_wfmod mbs (map fst oenvb).

  Variable DJab: forall lbl (IN: In lbl (map fst oenva)), ~ In lbl (map fst oenvb).
  Variable DJa: forall lbl (IN: In lbl (map fst oenva)), ~ In lbl (map fst ienv).
  
 
  Variable names: typename -> Prop.
  Variable CI: countably_infinite_set names.

  Definition namesa: typename -> Prop := 
    proj1_sig (cid _ (countably_infinite_set_split CI)).

  Definition namesb: typename -> Prop := 
    proj1_sig (cid _ (proj2_sig (cid _ (countably_infinite_set_split CI)))).

  Lemma names_split:
    <<CIa: countably_infinite_set namesa>> /\
    <<CIb: countably_infinite_set namesb>> /\
    <<UNN: names = nunion namesa namesb>> /\
    <<DJ: ndisj namesa namesb>>.
  Proof.
    apply (proj2_sig (cid _ (proj2_sig (cid _ (countably_infinite_set_split CI))))).
  Qed.


  Variable ancht: lang_spec.anchor tgt_lsi.
  Variable subt: list (label * word).
  Variable cigt: lang_spec.conf tgt_lsi.
  Variable cilt: lang_spec.conf tgt_lsi.

  Variable anchs: lang_spec.anchor src_lsi.
  Variable subs: list (label * src_val).
  Variable cigs: lang_spec.conf src_lsi.
  Variable cils: lang_spec.conf src_lsi.

  Definition lsegst := (tgt_link mat mbt) ancht (map snd subt).

  Definition anchat := ancht.
  Definition subat := subt.
  Definition lsegsat := mat anchat (map snd subat).

  Definition anchas := tt.
  Definition subas := subs.

  Definition anchbt := anchat + msize (map (snd <*> snd) lsegsat).
  Definition subbt := subt ++ map (fun p => (fst p, fst (snd p))) lsegsat.
  Definition lsegsbt := mbt anchbt (map snd subbt).

  Definition anchbs := tt.
  Definition subbs := subs ++ mas.


  Variable LMs: lang_spec.cload src_lsi (src_link mas mbs) anchs subs cigs cils.
  Variable LMt: lang_spec.cload tgt_lsi (tgt_link mat mbt) ancht subt cigt cilt.

  Definition cigas: lang_spec.conf src_lsi :=
    (pmemp, Just (subs ++ mas ++ mbs ++ proj1_sig (cid _ LMs)), Nothing).

  Definition cilas: lang_spec.conf src_lsi :=
    (pmemp, Nothing, Nothing).

  Definition cigat: lang_spec.conf tgt_lsi :=
    {| cpc := Nothing
     ; crf := cigt.(crf)
     ; cst := cigt.(cst)
     ; chp := proj1_sig (cid _ LMt) ## loadgh (map snd lsegsat)
    |}.

  Definition cilat: lang_spec.conf tgt_lsi :=
    {| cpc := Nothing
     ; crf := Nothing
     ; cst := pmemp
     ; chp := loadlh (map snd lsegsat)
    |}.


  Lemma lsegst_split: lsegst = lsegsat ++ lsegsbt.
  Proof.
    unfold lsegsbt, subbt, anchbt, lsegsat, lsegst, tgt_link, link, anchat, subat.
    rewrite map_app, map_map; auto.
  Qed.

  Lemma loadgh_split: loadgh (map snd lsegst) =
                      loadgh (map snd lsegsat) ## loadgh (map snd lsegsbt).
  Proof.
    rewrite <-loadgh_app, <-map_app, lsegst_split; auto.
  Qed.

  Lemma loadlh_split: loadlh (map snd lsegst) =
                      loadlh (map snd lsegsat) ## loadlh (map snd lsegsbt).
  Proof.
    rewrite <-loadlh_app, <-map_app, lsegst_split; auto.
  Qed.


  Lemma valid_all: pmvalid (proj1_sig (cid _ LMt) ##
                            loadgh (map snd lsegst) ## loadlh (map snd lsegst)).
  Proof.
    set (LMt' := cid _ LMt); destruct LMt' as [hp_imports X]; des; s.
    fold lsegst in GHP; rewrite GHP in GDJ; fold lsegst in LHP; rewrite LHP in LDJ.
    apply pmvalid_plus_plus; eauto; eapply loadgh_loadlh_valid, tgt_link_wfmod; eauto.
  Qed.

  Lemma finite_all: pmfinite (proj1_sig (cid _ LMt) ##
                              loadgh (map snd lsegst) ## loadlh (map snd lsegst)).
  Proof.
    assert (FIN:=LMt); apply tgt_cload_finite in FIN.
    set (LMt' := cid _ LMt); destruct LMt' as [hp_imports X]; des; s.
    fold lsegst in GHP; rewrite GHP in GDJ; fold lsegst in LHP; rewrite LHP in LDJ.
    rewrite GHP in FIN; rewrite LHP in FIN0; apply pmfinite_plusD in FIN; auto; des.
    apply pmfinite_plus; try apply pmfinite_plus; auto.
  Qed.
    

  Local Ltac validity VLD :=
     rewrite loadgh_split, loadlh_split in VLD; decomp_pmvalid VLD.

  Local Ltac prepare :=
    let X := fresh in let Y := fresh in
    assert (X:=valid_all); revert X; assert (X:=finite_all); revert X;
    set (X := cid _ LMt); destruct X as [hp_imports X]; des;
    set (Y := cid _ LMs); destruct Y as [envf Y]; des;
    intros FIN VLD; s in VLD; s in FIN.


  Lemma LMat: lang_spec.cload tgt_lsi mat anchat subat cigat cilat.
  Proof.
    unfold cigat; prepare; validity VLD.
    i; exists hp_imports; esplits; eauto.
  Qed.

  Lemma LMas: lang_spec.cload src_lsi mas anchas subas cigas cilas.
  Proof.
    unfold cigas, cilas; prepare; s; rr; esplits; eauto.
  Qed.


  Definition cigbs: lang_spec.conf src_lsi :=
    (pmemp, Just (subbs ++ mbs ++ proj1_sig (cid _ LMs)), Nothing).

  Definition cilbs: lang_spec.conf src_lsi :=
    (pmemp, Nothing, Nothing).

  Definition cigbt: lang_spec.conf tgt_lsi :=
    {| cpc := Nothing
     ; crf := cigt.(crf)
     ; cst := cigt.(cst)
     ; chp := proj1_sig (cid _ LMt) ## loadgh (map snd lsegsat)
                                    ## loadgh (map snd lsegsbt)
    |}.

  Definition cilbt: lang_spec.conf tgt_lsi :=
    {| cpc := Nothing
     ; crf := Nothing
     ; cst := pmemp
     ; chp := loadlh (map snd lsegsbt)
    |}.


  Lemma loadgh_in: forall loc (lsegs: list (label * segment))
      (IN: In loc (map (fst <*> snd) lsegs))
      (VLD: pmvalid (loadgh (map snd lsegs))),
    exists v, pmread (loadgh (map snd lsegs)) loc v.
  Proof.
    i; generalize dependent lsegs.
    induction lsegs; vauto; i; destruct a as (?&?&?); ss; des.
    - subst; eexists; econs; eauto.
    - edestruct IHlsegs; eauto using pmvalid_plus_r.
      econs; eapply pmread_plus; eauto using pmplus_comm.
  Qed.

  Lemma loadgh_out: forall loc (lsegs: list (label * segment)) v
      (IN: pmread (loadgh (map snd lsegs)) loc v),
    In loc (map (fst <*> snd) lsegs).
  Proof.
    i; induction lsegs; i; [apply pmread_empD in IN; auto|].
    destruct a as (?&?&?); ss; apply pmread_plusD in IN.
    des; [left|right; eauto]; apply pmread_cellD in IN; des; auto.
  Qed.


  Lemma LMbt: lang_spec.cload tgt_lsi mbt anchbt subbt cigbt cilbt.
  Proof.
    unfold cigbt; prepare; validity VLD.
    exists (hp_imports ## loadgh (map snd lsegsat)).
    esplits; s; eauto.
    - rewrite <-pmplus_assoc; eauto.
    - apply pmvalid_plus_plus; eauto.
    - i; split; intro IN.
      + unfold subbt in IN; rewrite map_app in IN; apply in_app_or in IN; des.
        * eapply GDOM in IN; des; eauto; econs; eapply pmread_plus; eauto.
        * edestruct loadgh_in; eauto using pmvalid_plus_l
          ; [|by econs; eapply pmread_plus; eauto using pmplus_comm]
          ; rewrite map_map in IN; auto.
      + unfold subbt; rewrite map_app; apply in_or_app.
        des; apply pmread_plusD in IN; eauto; des; [by left; eapply GDOM; eauto|].
        right; rewrite map_map; eapply loadgh_out; eauto.
    - rewrite pmplus_comm; apply pmvalid_plus_plus; [M| |]
      ; (Mdo rewrite pmplus_comm); auto.
  Qed.

  Lemma LMbs: lang_spec.cload src_lsi mbs anchbs subbs cigbs cilbs.
  Proof.
    unfold cigbs, cilbs; prepare; s; rr; esplits; eauto.
  Qed.


  Lemma oenva_mat: map fst oenva = map fst lsegsat.
  Proof.
    assert (X:=WFat); rrdes X; specialize (OUT anchat (map snd subat)); auto.
  Qed.

  Lemma oenva_mas: map fst oenva = map fst mas.
  Proof.
    assert (X:=WFas); rrdes X; auto.
  Qed.

  Lemma oenvb_mbt: map fst oenvb = map fst lsegsbt.
  Proof.
    assert (X:=WFbt); rrdes X; specialize (OUT anchbt (map snd subbt)); auto.
  Qed.

  Lemma oenvb_mbs: map fst oenvb = map fst mbs.
  Proof.
    assert (X:=WFbs); rrdes X; auto.
  Qed.

  Lemma oenva_mbt: forall f ty vt code (IN: lookup f oenva = Some ty),
    ~ In (f, (vt, code)) lsegsbt.
  Proof.
    ii; apply list_assoc_someD in IN.
    cut (In f (map fst oenva)); [i|apply in_map_iff; esplits; try apply IN; auto].
    apply DJab in H; vauto; erewrite oenvb_mbt; apply in_map_iff.
    esplits; [|eauto]; eauto.
  Qed.

  Lemma oenvb_mat: forall f ty vt code (IN: lookup f oenvb = Some ty),
    ~ In (f, (vt, code)) lsegsat.
  Proof.
    ii; apply list_assoc_someD in IN.
    cut (In f (map fst oenvb)); [i|apply in_map_iff; esplits; try apply IN; auto].
    apply DJab in H; vauto; erewrite oenva_mat; apply in_map_iff.
    esplits; [|eauto]; eauto.
  Qed.

  Lemma oenva_mbs: forall f ty v (IN: lookup f oenva = Some ty),
    ~ lookup f mbs = Some v.
  Proof.
    ii; apply list_assoc_someD in IN.
    cut (In f (map fst oenva)); [i|apply in_map_iff; esplits; try apply IN; auto].
    apply DJab in H; vauto; erewrite oenvb_mbs; apply in_map_iff.
    esplits; [|eapply list_assoc_someD]; s; eauto.
  Qed.

  Lemma oenvb_mas: forall f ty v (IN: lookup f oenvb = Some ty),
    ~ lookup f mas = Some v.
  Proof.
    ii; apply list_assoc_someD in IN.
    cut (In f (map fst oenvb)); [i|apply in_map_iff; esplits; try apply IN; auto].
    apply DJab in H; vauto; erewrite oenva_mas; apply in_map_iff.
    esplits; [|eapply list_assoc_someD]; s; eauto.
  Qed.

  Lemma ienv_mat: forall f ty v (IN: lookup f ienv = Some ty),
    ~ lookup f (map (fun p => (fst p, fst (snd p))) lsegsat) = Some v.
  Proof.
    i; apply list_assoc_someD, (in_map fst) in FH; rewrite map_map in FH.
    s in FH; rewrite <-(eta fst) in FH; fold segment in FH. 
    rewrite <-oenva_mat in FH; eapply DJa; eauto.
    apply in_map_iff; esplits; [|eapply list_assoc_someD]; s; eauto.
  Qed.

  Lemma ienv_mas: forall f ty v (IN: lookup f ienv = Some ty),
    ~ lookup f mas = Some v.
  Proof.
    i; apply list_assoc_someD, (in_map fst) in FH.
    s in FH; rewrite <-oenva_mas in FH; eapply DJa; eauto.
    apply in_map_iff; esplits; [|eapply list_assoc_someD]; s; eauto.
  Qed.


  Variable SUBt: map fst subt = map fst ienv.
  Variable SUBs: map fst subs = map fst ienv.

  Lemma SUBat: map fst subat = map fst ienv.
  Proof.
    auto.
  Qed.

  Lemma SUBas: map fst subas = map fst ienv.
  Proof.
    auto.
  Qed.

  Lemma SUBbt: map fst subbt = map fst (ienv ++ oenva).
  Proof.
    unfold subbt, lsegsat; rewrite !map_app, SUBt, map_map; s; rewrite oenva_mat; auto.
  Qed.

  Lemma SUBbs: map fst subbs = map fst (ienv ++ oenva).
  Proof.
    unfold subbs; rewrite !map_app, SUBs, oenva_mas; auto.
  Qed.
 

  Variable lwa: world_loca ts_gw.
  Variable sia: (wlift lwa).(state).
  Variable NAMa: forall tn, supp lwa tn -> namesa tn.
  Variable STBa: stable lwa.
  Variable INIga: forall g (GK: gknow (wlift lwa) g),
    ts_gw.(confrel) (fun sg => g (sg, snd sia)) (fst sia) cigat cigas.
  Variable INIla: forall g (GK: gknow (wlift lwa) g),
    lwa.(confrel) g sia cilat cilas.
  Variable REFa: forall ty lt ls, ~ ts_gw.(rquery) (fst sia) ty lt ls.

  Variable lwb: world_loca ts_gw.
  Variable sib: state (wlift lwb).
  Variable NAMb: forall tn, supp lwb tn -> namesb tn.
  Variable STBb: stable lwb.
  Variable INIgb: forall g (GK: gknow (wlift lwb) g),
    ts_gw.(confrel) (fun sg => g (sg, snd sib)) (fst sib) cigbt cigbs.
  Variable INIlb: forall g (GK: gknow (wlift lwb) g),
    lwb.(confrel) g sib cilbt cilbs.
  Variable REFb: forall ty lt ls, ~ ts_gw.(rquery) (fst sib) ty lt ls.


  Definition lw : world_loca ts_gw :=
    wprod lwa lwb.

  Definition sig: ts_gw.(state) :=
    let '((siat, sias, _), _) := sia in
    let '((sibt, sibs, _), _) := sib in
    ((fst sibt, snd siat \2/ snd sibt), sibs, refst_emp).


  Lemma read_cigat_cigbt: forall l v
        (RD: pmread cigat.(chp) l v),
      pmread cigbt.(chp) l v.
  Proof.
    unfold cigat, cigbt; prepare; s; i.
    rewrite pmplus_assoc; eapply pmread_plus; eauto.
    validity VLD; rewrite <-pmplus_assoc; apply pmvalid_plus_plus; auto.
  Qed.


  Lemma cigt_cigbt: cigt = cigbt.
  Proof.
    unfold cigbt; prepare; clear INIga INIgb LMt; destruct cigt; clarify; ss; f_equal; auto.
    rewrite GHP; fold lsegst.
    rewrite pmplus_assoc, lsegst_split, map_app, loadgh_app, <-pmplus_assoc; auto.
  Qed.

  Lemma cigs_cigbs: cigs = cigbs.
  Proof.
    unfold cigbs; prepare; rewrite Ecg; s.
    unfold src_link; rewrite <-app_assoc, app_assoc; do 4 f_equal.
  Qed.


  Lemma NOREFa: bot3 = snd (fst sia).
  Proof.
    clear -REFa; destruct sia as [[]]; extensionalities ty lt ls; apply prop_ext.
    split; i; vauto; apply (@REFa ty lt (Vloc ls)); exists ls; eauto. 
  Qed.

  Lemma NOREFb: bot3 = snd (fst sib).
  Proof.
    clear -REFb; destruct sib as [[]]; extensionalities ty lt ls; apply prop_ext.
    split; i; vauto; apply (@REFb ty lt (Vloc ls)); exists ls; eauto.
  Qed.


  Variable DJN: forall tn (INa: lwa.(supp) tn) (INb: lwb.(supp) tn), False.
  


  Lemma INIg: forall g
      (GK: gknow (wlift lw) g),
    ts_gw.(confrel) (fun sg => g (sg, (snd sia, snd sib))) sig cigt cigs.
  Proof.
    unfold sig; i; assert (NOREFa:=NOREFa); assert (NOREFb:=NOREFb).
    destruct sia as [[[siat sias] siaref] sial].
    destruct sib as [[[sibt sibs] sibref] sibl].
    assert (X:=LMt); rrdes X.
    edestruct (gknow_prod STBa STBb) as (GKa & GKb); eauto.
    exploit INIga; [by eapply (GKa sibl)|]; intro CRa.
    exploit INIgb; [by eapply (GKb sial)|]; intro CRb.
    rrdes CRa CRb; rrdes WHX WHX0; rrdes WHY WHY0.
    assert (EMP: ha' = pmemp /\ hb' = pmemp /\ ha'0 = pmemp /\ hb'0 = pmemp)
    ; [|clear RELV RELV0; des; subst].
    { splits; apply pmemp_from_read; eauto; i; clean.
      - rewrite <-NOREFa in DOMa; exploit DOMa; eauto; i; des; vauto.
      - rewrite <-NOREFa in DOMb; exploit DOMb; eauto; i; des; vauto.
      - rewrite <-NOREFb in DOMa0; exploit DOMa0; eauto; i; des; vauto.
      - rewrite <-NOREFb in DOMb0; exploit DOMb0; eauto; i; des; vauto.
    }
    rrdes CPa CPb CPa0 CPb0; ss; rr; esplits.
    - rr; splits; s; rr; esplits; eauto.
      s; i; des; [|eapply HP0]; eauto; eapply (@repr_mon (chp hax) (chp hax0)); eauto; i.
      rewrite (@cplus_comm tgt_lsi), cplus_emp in HPA. 
      
      rewrite (@cplus_comm tgt_lsi), cplus_emp in HPA0.
      rewrite HPA0; rewrite HPA in IN0; apply read_cigat_cigbt; auto.
    - rr; esplits; eauto; [| |by i; ss; des; vauto]
      ; i; apply pmread_empD in ALC; vauto.
    - rewrite cigt_cigbt; auto.
    - rewrite cigs_cigbs; auto.
  Qed.


  Lemma cils_link: src_cplus cilas cilbs = cils.
  Proof.
    unfold cilas, cilbs, src_cplus; assert (X:=LMs); rrdes X; s; noemp; auto.
  Qed.

  Lemma cilt_link: tgt_cplus cilat cilbt = cilt.
  Proof.
    unfold cilat, cilbt, tgt_cplus; assert (X:=LMt); rrdes X; s; noemp.
    s; rewrite <-LPC, <-LRF, <-LST, <-loadlh_app, <-map_app, <-lsegst_split.
    unfold lsegst; rewrite <-LHP; clear -LMt; destruct cilt; f_equal; eauto.
  Qed.


  Definition tgt_dummy: tgt_conf := mkconf (Just 42) Nothing pmemp pmemp.
  Definition src_dummy: src_conf := (pmemp, Nothing, Just Eunit).


  Lemma cst_cigt: pmvalid (cst cigt).
  Proof.
    prepare; destruct (cst cigt); vauto; specialize (GST0 0); rrdes GST0; auto.
  Qed.


  Lemma realat: exists maat, real tgt_lsi (cilat # cigt # tgt_dummy) maat.
  Proof.
    s; unfold cilat, tgt_real; prepare.
    exists {| mpc := 42; mrf := rf; mst := cigt.(cst)
            ; mhp := loadlh (map snd lsegsat) ## chp cigt ## pmemp |}.
    clear INIga INIgb LMt; destruct cigt; unfold tgt_dummy; rr; ss. 
    rewrite GPC, GRF, GHP; esplits; s; noemp; eauto using cst_cigt.
    - validity VLD; fold lsegst; rewrite loadgh_split; apply pmvalid_plus_plus; [M| |]
      ; (Mdo rewrite pmplus_comm); eauto using pmvalid_plus_plus.
    - fold lsegst; split_pmfinite FIN VLD.
      apply pmfinite_plus; do 2 (try apply pmfinite_plus); auto using pmfinite_emp.
      rewrite loadgh_split, loadlh_split in *.
      split_pmfinite FIN0 VLD0; split_pmfinite FIN1 VLD1; auto. 
  Grab Existential Variables.
  exact 0.
  Qed.

  Lemma realbt: exists mabt, real tgt_lsi (cilbt # cigt # tgt_dummy) mabt.
  Proof.
    s; unfold cilat, tgt_real; prepare.
    exists {| mpc := 42; mrf := rf; mst := cigt.(cst)
            ; mhp := loadlh (map snd lsegsbt) ## chp cigt ## pmemp |}.
    clear INIga INIgb LMt; destruct cigt; unfold tgt_dummy; rr; ss. 
    rewrite GPC, GRF, GHP; esplits; s; noemp; eauto using cst_cigt.
    - validity VLD; fold lsegst; rewrite loadgh_split; apply pmvalid_plus_plus; [M| |]
      ; (Mdo rewrite pmplus_comm); eauto using pmvalid_plus_plus.
    - fold lsegst; split_pmfinite FIN VLD.
      apply pmfinite_plus; do 2 (try apply pmfinite_plus); auto using pmfinite_emp.
      rewrite loadgh_split, loadlh_split in *.
      split_pmfinite FIN0 VLD0; split_pmfinite FIN1 VLD1; auto. 
  Grab Existential Variables.
  exact 0.
  Qed.

  Lemma realas: exists maas, real src_lsi (cilas # cigs # src_dummy) maas.
  Proof.
    s; unfold cilas, src_cplus, src_real; s; assert (X:=LMs); rrdes X; s.
    exists (pmemp, subs ++ src_link mas mbs ++ envf, Eunit).
    rewrite Ecg; esplits; eauto using pmvalid_emp, pmfinite_emp; s; noemp; auto.
  Qed.

  Lemma realbs: exists mabs, real src_lsi (cilbs # cigs # src_dummy) mabs.
  Proof.
    s; unfold cilas, src_cplus, src_real; assert (X:=LMs); rrdes X; s.
    exists (pmemp, subs ++ src_link mas mbs ++ envf, Eunit).
    rewrite Ecg; esplits; eauto using pmvalid_emp, pmfinite_emp; s; noemp; auto.
  Qed.


  Lemma cigas_cigbs: cigas = cigbs.
  Proof.
    unfold cigas, cigbs.
    do 3 f_equal; rewrite app_assoc; f_equal.
  Qed.


  Lemma sias_sibs: forall g
      (GK: gknow (wlift lw) g),
   snd (fst (fst sia)) = snd (fst (fst sib)).
  Proof.
    unfold sig; i; assert (NOREFa:=NOREFa); assert (NOREFb:=NOREFb).
    destruct sia as [[[siat sias] siaref] sial].
    destruct sib as [[[sibt sibs] sibref] sibl].
    assert (X:=LMt); rrdes X.
    edestruct (gknow_prod STBa STBb) as (GKa & GKb); eauto.
    exploit INIga; [by eapply (GKa sibl)|]; intro CRa.
    exploit INIgb; [by eapply (GKb sial)|]; intro CRb.
    rrdes CRa CRb; rrdes WHX WHX0; rrdes WHY WHY0.
    assert (EMP: ha' = pmemp /\ hb' = pmemp /\ ha'0 = pmemp /\ hb'0 = pmemp)
    ; [|clear RELV RELV0; des; subst].
    { splits; apply pmemp_from_read; eauto; i; clean.
      - rewrite <-NOREFa in DOMa; exploit DOMa; eauto; i; des; vauto.
      - rewrite <-NOREFa in DOMb; exploit DOMb; eauto; i; des; vauto.
      - rewrite <-NOREFb in DOMa0; exploit DOMa0; eauto; i; des; vauto.
      - rewrite <-NOREFb in DOMb0; exploit DOMb0; eauto; i; des; vauto. }
    rewrite cigas_cigbs, <-HPB0 in HPB; s in HPB.
    
    
    destruct hbx, hbx0; ss; destruct p, e; vauto; destruct h, e1; vauto.
    destruct p0; ss; vauto; rrdes CPb CPb0; clarify; s in H1; clarify.
  Qed.


  Lemma INIla': exists s, forall g (GK: gknow (wlift lw) g),
    extpub _ s (snd sia) /\
    lwa.(confrel) (prod_gka g (snd sib)) (sig, s) cilat cilas.
  Proof.
    assert (sias_sibs := sias_sibs).

    cut (forall g (GK: gknow (wlift lw) g), exists s,
           extpub _ s (snd sia) /\
           lwa.(confrel) (prod_gka g (snd sib)) (sig, s) cilat cilas).
    { i; exploit H; eauto using gkmin_bot_gknow; i; des.
      exists s; i; split; auto.
      eapply confrel_mon; eauto.
      i; eapply prod_gka_mon; eauto; i; eapply gkmin_min; last first; vauto.
    }

    i; specialize (sias_sibs _ GK).
    edestruct (gknow_prod STBa STBb) as (GKa & GKb); eauto.
    assert (INIg := INIg).
    assert (REALas:=realas); rrdes REALas; eauto.
    assert (REALat:=realat); rrdes REALat; eauto.

    cut (exists s: state lwa,
           extpub lwa s (snd sia) /\
           confrel lwa (prod_gka g (snd sib)) (sig, s) cilat cilas); eauto.
    unfold sig in *.
    i; destruct sia as [[[]]], sib as [[[]]].
    eapply STBa; eauto.
    - repeat split; s; eauto.
      i; eapply REFa; s; rr; eauto.
    - hexploit INIg; eauto; []; i.  econs.
      rewrite (cplus_comm cigs), (cplus_swap _ cilas), (cplus_comm cilas) in REALas.
      rewrite (cplus_comm cigt), (cplus_swap _ cilat), (cplus_comm cilat) in REALat.
      esplits; econs; first 1 last; first last.
      + eapply confrel_mon; eauto.
      + instantiate (2 := cemp); noemp; eauto.
      + instantiate (2 := cemp); noemp; eauto.
      + unfold tgt_dummy, src_dummy; split; rr; s; eauto.
  Qed.

  Lemma INIlb': exists s, forall g (GK: gknow (wlift lw) g),
    extpub _ s (snd sib) /\
    lwb.(confrel) (prod_gkb g (snd sia)) (sig, s) cilbt cilbs.
  Proof.
    cut (forall g (GK: gknow (wlift lw) g), exists s,
           extpub _ s (snd sib) /\
           lwb.(confrel) (prod_gkb g (snd sia)) (sig, s) cilbt cilbs).
    { i; exploit H; eauto using gkmin_bot_gknow; i; des.
      exists s; i; split; auto.
      eapply confrel_mon; eauto.
      i; eapply prod_gkb_mon; eauto; i; eapply gkmin_min; last first; vauto.
    }

    i; edestruct (gknow_prod STBa STBb) as (GKa & GKb); eauto.
    assert (INIg := INIg).
    assert (REALbs:=realbs); rrdes REALbs; eauto.
    assert (REALbt:=realbt); rrdes REALbt; eauto.

    cut (exists s : state lwb,
           extpub lwb s (snd sib) /\
           confrel lwb (prod_gkb g (snd sia)) (sig, s) cilbt cilbs); eauto.
    unfold sig in *.
    i; destruct sia as [[[]]], sib as [[[]]].
    eapply STBb; eauto.
    - repeat split; s; eauto.
      i; eapply REFb; s; rr; eauto.
    - hexploit INIg; eauto; []; i.  econs.
      rewrite (cplus_comm cigs), (cplus_swap _ cilbs), (cplus_comm cilbs) in REALbs.
      rewrite (cplus_comm cigt), (cplus_swap _ cilbt), (cplus_comm cilbt) in REALbt.
      esplits; econs; first 1 last; first last.
      + eapply confrel_mon; eauto.
      + instantiate (2 := cemp); noemp; eauto.
      + instantiate (2 := cemp); noemp; eauto.
      + unfold tgt_dummy, src_dummy; split; rr; s; eauto.
  Qed.


  Definition sil: lw.(state) :=
    (proj1_sig (cid _ INIla'), proj1_sig (cid _ INIlb')).

  Lemma EXT_sil: extpub _ sil (snd sia, snd sib).
  Proof.
    unfold sig, sil in *.
    set (Xa := cid _ INIla'); destruct Xa as [sial' Xa]; des.
    set (Xb := cid _ INIlb'); destruct Xb as [sibl' Xb]; des.
    s; unfold sig in *; clean.
    specialize (Xa _ (gkmin_bot_gknow _)); des.
    specialize (Xb _ (gkmin_bot_gknow _)); des.
    split; eauto.
  Qed.

  Lemma INIl: forall g
      (GK: gknow (wlift lw) g),
    lw.(confrel) g (sig, sil) cilt cils.
  Proof.
    unfold sig, sil in *.
    set (Xa := cid _ INIla'); destruct Xa as [sial' Xa]; des.
    set (Xb := cid _ INIlb'); destruct Xb as [sibl' Xb]; des.
    s; unfold sig in *; clean.
    assert (INIg:=INIg); assert (sias_sibs:=sias_sibs); i.
    destruct sia as [[[siat sias] siaref] sial].
    destruct sib as [[[sibt sibs] sibref] sibl].
    assert (X:=LMt); rrdes X.
    edestruct (gknow_prod STBa STBb) as (GKa & GKb); eauto.
    exploit Xa; [by eauto|]; intro CRa; des.
    exploit Xb; [by eauto|]; intro CRb; des.
    clean.
    rr; ss; esplits; eauto 10 using cilt_link, cils_link.
    - eapply confrel_mon; eauto using (@prod_gka_mon tgt_lsi src_lsi ts_gw).
    - eapply confrel_mon; eauto using (@prod_gkb_mon tgt_lsi src_lsi ts_gw).
  Qed.


  Lemma REF: forall ty lt ls,
    ~ ts_gw.(rquery) sig ty lt ls.
  Proof.
    unfold sig; destruct sia as [[]], sib as [[]]; destruct s, s2; s; i.
    inversion FH; des; ss.
  Qed.


  Lemma vload_at:
    forall f ty (IN: lookup f oenva = Some ty)
           vt (Et: vload tgt_lsi (tgt_link mat mbt) ancht subt f vt),
      vload tgt_lsi mat anchat subat f vt.
  Proof.
    i; rrdes Et; econs; fold lsegst in Et; rewrite lsegst_split in Et.
    eapply list_assoc_app_or in Et; des; eauto; exfalso; apply (DJab f).
    - apply list_assoc_someD in IN.
      eapply in_map_iff; esplits; eauto; s; auto.
    - apply list_assoc_someD in Et0; erewrite oenvb_mbt.
      eapply in_map_iff; esplits; eauto; s; auto.
  Qed.

  Lemma vload_bt:
    forall f ty (IN: lookup f oenvb = Some ty)
           vt (Et: vload tgt_lsi (tgt_link mat mbt) ancht subt f vt),
      vload tgt_lsi mbt anchbt subbt f vt.
  Proof.
    i; rrdes Et; econs; fold lsegst in Et; rewrite lsegst_split in Et.
    eapply list_assoc_app_or in Et; des; eauto; exfalso; apply (DJab f).
    - apply list_assoc_someD in Et; rewrite oenva_mat.
      apply in_map_iff; esplits; eauto; s; auto.
    - apply list_assoc_someD in IN.
      apply in_map_iff; esplits; eauto; s; auto.
  Qed.

  Lemma vload_as:
    forall f ty (IN: lookup f oenva = Some ty)
           vs (Et: vload src_lsi (src_link mas mbs) anchs subs f vs),
      vload src_lsi mas anchas subas f vs.
  Proof.
    unfold src_vload; i; des; s in Et.
    apply list_assoc_app_or in Et; des; eauto; exfalso; apply (DJab f).
    - eapply list_assoc_someD in IN; eapply in_map_iff; econs; split; eauto; s; auto.
    - apply list_assoc_someD in Et0; erewrite oenvb_mbs; eapply in_map_iff; esplits; eauto; s; auto.
  Qed.

  Lemma vload_bs:
    forall f ty (IN: lookup f oenvb = Some ty)
           vs (Et: vload src_lsi (src_link mas mbs) anchs subs f vs),
      vload src_lsi mbs anchbs subbs f vs.
  Proof.
    unfold src_vload; i; des; s in Et.
    apply list_assoc_app_or in Et; des; eauto; exfalso; apply (DJab f).
    - apply list_assoc_someD in Et; erewrite oenva_mas; eapply in_map_iff; esplits; eauto; s; auto.
    - eapply list_assoc_someD in IN; eapply in_map_iff; econs; split; eauto; s; auto.
  Qed.


  Variable SIMa: forall f ty (IN: lookup f oenva = Some ty), exists vt vs,
                   <<Et: vload tgt_lsi mat anchat subat f vt>> /\
                   <<Es: vload src_lsi mas anchas subas f vs>> /\
                   <<SIM: @tlsim _ _ ts_gw lwa ienv subat subas sia vt vs ty>>.

  Variable SIMb: forall f ty (IN: lookup f oenvb = Some ty), exists vt vs,
                   <<Et: vload tgt_lsi mbt anchbt subbt f vt>> /\
                   <<Es: vload src_lsi mbs anchbs subbs f vs>> /\
                   <<SIM: @tlsim _ _ ts_gw lwb (ienv ++ oenva) subbt subbs sib vt vs ty>>.


  Lemma oenva_subt: forall f ty v
      (INoenva: lookup f oenva = Some ty)
      (INsubt: lookup f subt = Some v),
    False.
  Proof. i.
    apply list_assoc_someD, (in_map fst) in INsubt; rewrite SUBt in INsubt.
    apply list_assoc_someD, (in_map fst) in INoenva; rewrite oenva_mat in INoenva.
    eapply (list_assoc_map eqdec) in INsubt; apply (list_assoc_map eqdec) in INoenva; des.
    eapply ienv_mat with (v := fst y0); eauto; induction lsegsat; vauto.
    destruct y0, a as (?&?&?); s; des_if; subst; ss; [|apply IHl]; des_ifH INoenva; vauto.
  Qed.

  Lemma oenva_subs: forall f ty v
      (INoenva: lookup f oenva = Some ty)
      (INsubs: lookup f subs = Some v),
    False.
  Proof. i.
    apply list_assoc_someD, (in_map fst) in INsubs; rewrite SUBs in INsubs.
    apply list_assoc_someD, (in_map fst) in INoenva; rewrite oenva_mas in INoenva.
    eapply (list_assoc_map eqdec) in INsubs; apply (list_assoc_map eqdec) in INoenva; des.
    eapply ienv_mas; eauto.
  Qed.


  Lemma SIM_ext: forall sg sa sb g
      (GK: gknow (wlift lw) g)
      (EXT: ext (wlift lw) (sg, (sa, sb)) (sig, sil)),
    <<EXTa: ext (wlift lwa) (sg, sa) sia>> /\
    <<EXTb: ext (wlift lwb) (sg, sb) sib>>.
  Proof.
    unfold sig, sil in *.
    set (Xa := cid _ INIla'); destruct Xa as [sial' Xa]; des.
    set (Xb := cid _ INIlb'); destruct Xb as [sibl' Xb]; des.
    unfold sig in *; simpl proj1_sig; clean.
    i; assert (EXT':=EXT); assert (sias_sibs:=sias_sibs).
    hexploit Xa; [by eauto|]; i; des.
    hexploit Xb; [by eauto|]; i; des.
    rrdes EXT' EXT'0; unfold sig in *.
    destruct sia as [[[]]], sib as [[[]]]; destruct s3, s.
    splits; split; auto; clean.
    - split; [split|]; s; eauto
      ; try (by i; destruct sg as [[]]; ss; des; apply EXT; eauto)
      ; try (by i; eapply EXT, REFa; rr; eauto)
      ; try (by i; eapply EXT, REFb; rr; eauto).
      ss; des; erewrite sias_sibs; eauto.
    - eauto using ext_trans.
    - split; [split|]; s; eauto
      ; try (by i; destruct sg as [[]]; ss; des; apply EXT; eauto)
      ; try (by i; eapply EXT, REFa; rr; eauto)
      ; try (by i; eapply EXT, REFb; rr; eauto).
      ss; des; eauto.
    - eauto using ext_trans.
  Qed.


  Lemma SIM: forall f ty (IN: lookup f (oenva ++ oenvb) = Some ty), exists vt vs,
               <<Et: vload tgt_lsi (tgt_link mat mbt) ancht subt f vt>> /\
               <<Es: vload src_lsi (src_link mas mbs) anchs subs f vs>> /\
               <<SIM: @tlsim _ _ ts_gw lw ienv subt subs (sig, sil) vt vs ty>>.
  Proof.
    ii; dup SIMa; dup SIMb.
    specialize (@SIMa f ty); specialize (@SIMb f ty).
    apply list_assoc_app_or in IN; des.
    - hexploit SIMa0; eauto; []; i; des; esplits.
      + rr; fold lsegst; rewrite lsegst_split.
        rrdes Et; fold lsegsat in Et; rewrite list_assoc_app, Et; eauto.
      + rr; rrdes Es; rewrite list_assoc_app, Es; eauto. 
      +                         
        ii.
        edestruct (gknow_prod STBa STBb) as (GKa & GKb); eauto.
        destruct s as (sg & sa & sb).
        exploit SIM; first eapply SIM_ext; eauto.
        * rrdes RLS RLS.
          
          rr; eauto.
        * i; dup RS; exploit RS; eauto; intro X; inversion X; eauto. 
        * intro SIMa''; inversion SIMa''; [econs 1|econs 2]
          ; eauto; invc VR; eapply gk_mon; eauto; split; s; eauto.
    - hexploit SIMb0; eauto; []; i; des; esplits.
      + rr; fold lsegst; rewrite lsegst_split.
        rrdes Et; fold lsegsbt in Et; rewrite list_assoc_app, Et.
        destruct (lookup f lsegsat) eqn:E.
        { exfalso.
          destruct s; eapply oenvb_mat; eauto using list_assoc_someD.
        }
        eauto.
      + rr; rrdes Es; rewrite list_assoc_app, Es; eauto. 
        destruct (lookup f mas) eqn:E.
        { exfalso.
          eapply oenvb_mas; eauto using list_assoc_someD.
        }
        eauto.
      +                         
        ii.
        edestruct (gknow_prod STBa STBb) as (GKa & GKb); eauto.
        destruct s as (sg & sa & sb).
        exploit SIM; first eapply SIM_ext; eauto.
        * rrdes RLS RLS.
          
          rr; eauto.
        * { i; apply list_assoc_app_or in INienv; des.
            - apply list_assoc_app_or in INsuba; des; [|by exfalso; eapply ienv_mat; eauto].
              apply list_assoc_app_or in INsubb; des; [|by exfalso; eapply ienv_mas; eauto].
              dup RS; exploit RS; eauto; intro SIMb''; inversion SIMb''; eauto.
            - apply list_assoc_app_or in INsuba; des; [by exfalso; eapply oenva_subt; eauto|].
              apply list_assoc_app_or in INsubb; des; [by exfalso; eapply oenva_subs; eauto|].
              eapply list_assoc_map_pairD in INsuba0.
              des; destruct z; clarify.
              exploit SIMa0; eauto; []; i; des.
Require Import ssr.
              rrdes Es0; rewrite Es0 in INsubb0; clarify.
              rrdes Et0; rewrite Et0 in INsuba0; clarify.
              exploit SIM0; first eapply SIM_ext; eauto.
              + rrdes RLS RLS; rr; eauto.
              + i; dup RS; exploit RS; eauto; intro X; inversion X; eauto.
              + intro SIMa''; inversion SIMa''; econs; auto.
                * invc VR; econs; eauto.
                  eapply gk_mon; eauto; split; s; eauto.
                * invc VR; econs; eauto.
                  eapply gk_mon; eauto; split; s; eauto.
          }
        * intro SIMb''; inversion SIMb''; [econs 1|econs 2]
          ; eauto; invc VR; eapply gk_mon; eauto; split; s; eauto.
  Qed.
  
End msim.


(** The linking theorem, a.k.a. horizontal compositionality. *)

Theorem linking: forall ienv mas mat oenva mbs mbt oenvb
    (SIMa: msim ts_gw ienv oenva mat mas)
    (SIMb: msim ts_gw (ienv ++ oenva) oenvb mbt mbs)
    (WFat: tgt_wfmod mat (map fst oenva))
    (WFas: src_wfmod mas (map fst oenva))
    (WFbt: tgt_wfmod mbt (map fst oenvb))
    (WFbs: src_wfmod mbs (map fst oenvb))
    (DJab: forall lbl (IN: In lbl (map fst oenva)), ~ In lbl (map fst oenvb))
    (DJa: forall lbl (IN: In lbl (map fst oenva)), ~ In lbl (map fst ienv)),
  msim ts_gw ienv (oenva ++ oenvb) (tgt_link mat mbt) (src_link mas mbs).
Proof.
  ii; assert (X:=names_split CI); des.
  specialize (SIMa _ CIa); specialize (SIMb _ CIb); rrdes SIMa SIMb.
  exists (lw lw0 lw1).
  intros ancht subt cigt cilt ? anchs subs cigs cils ? ? ?.
  exploit (SIMa (anchat ancht) (subat subt) (cigat LMa)
                (cilat mat ancht subt))
  ; eauto using LMat, LMas; i; des.
  exploit (SIMb (anchbt mat ancht subt) (subbt mat ancht subt) (cigbt LMa)
                (cilbt mat mbt ancht subt))
  ; eauto using LMbt, LMbs, SUBbt, SUBbs; i; des.
  assert (forall tn : typename, supp lw0 tn -> supp lw1 tn -> False).
  { i; apply NAM in H; apply NAM0 in H0; eauto. }
  instantiate (1:=LMb) in INIg0.
  instantiate (1:=LMb) in INIg1.
  set (sil := @sil mat mbt mas mbs oenva oenvb WFat WFbt DJab ancht subt cigt
                   cilt anchs subs cigs cils LMb LMa lw0 si STA INIg0 INIl0
                   REF0 lw1 si0 STA0 INIg1 INIl1 REF1 H). 
  exists (sig si si0, sil); splits.
  - s; i; rewrite UNN; des; [left|right]; eauto using NAM, NAM0.
  - eapply stable_prod; eauto.
  - i. clean. eapply confrel_mon. eapply (INIg oenva oenvb WFat WFbt); eauto.
    s; i; eapply (gk_mon GK); eauto.
    split; eauto using extpub_in_ext, EXT_sil.
  - i; eapply (INIl oenva oenvb WFat WFbt); eauto.
  - eapply REF; eauto.
  - eapply SIM; eauto.
Qed.                            


