(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List Program.
Require Import ssr countable vgtac basis lang_common lang_spec lang_mid lang_mid_renaming.
Require Import model_common model_untyped model_properties_untyped gw_untyped.
Require Import hcomp_esim_untyped.

Set Implicit Arguments.
Global Set Bullet Behavior "Strict Subproofs".


(** * Properties about [mm_gw]. *)


Lemma mm_confrel_shape: forall g (s: state mm_gwsts) ha hb
    (WH: confrel mm_gw g s ha hb),
  exists ha' hb',
    ha = (ha', Nothing) /\
    hb = (hb', Nothing).
Proof.
  i; rrdes WH WHX WHY; rrdes CPa CPb; subst; noemp; eauto.
Qed.


(** Inversion for untyped value closure. *)

Lemma mm_vclosDb: forall (w: world_loca mm_gw) s r va vb
    (VCL: vclos (wlift w) s r va vb),
  match vb with
    | IVunit =>
        va = IVunit
    | IVnat n =>
        va = IVnat n
    | IVpair vb' vb'' => exists va' va'',
        va = IVpair va' va'' /\
        <<VCL': vclos (wlift w) s r va' vb'>> /\
        <<VCL'': vclos (wlift w) s r va'' vb''>>
    | IVinl vb' => exists va',
        va = IVinl va' /\
        <<VCL': vclos (wlift w) s r va' vb'>>
    | IVinr vb' => exists va',
        va = IVinr va' /\
        <<VCL': vclos (wlift w) s r va' vb'>>
    | IVloc _ => exists va',
        va = IVloc va' /\
        <<QR: mm_gw.(rquery) (fst s) va vb>>
    | IVclo envb (IEfix xfb xvb xkb eb) => exists enva xfa xva xka ea,
        va = IVclo enva (IEfix xfa xva xka ea) /\
        <<RV: r va vb>>
    | IVclo envb (IEgen xkb eb) => exists enva xka ea,
        va = IVclo enva (IEgen xka ea) /\
        <<RV: r va vb>>
    | _ => True
  end.
Proof.
  i; destruct vb eqn:E; rewrite <-E in VCL
  ; move VCL at top; revert_until VCL; pattern va, vb
  ; try (by apply (vclos_induction_strong VCL); i
         ; invc IH; des; try inv CA CB; eauto
         ; [rrdes VR; esplits; rr|inv CB|inv VR CA CB]; vauto).
  apply (vclos_induction_strong VCL); i.
  subst; invc IH; des; try invc CA CB; auto.
    + specialize (VR _ _ eq_refl); destruct e; vauto.
    + specialize (VR _ _ eq_refl); destruct e; vauto.
    + rrdes VR; vauto.
    + inv CB.
    + invc VR CA CB; eauto 10.
Grab Existential Variables.
exact IVunit. exact IVunit. exact IVunit. exact IVunit.
Qed.


(** Convenvience lemma for reasoning about [IEread]. *)

Lemma mm_refconfrel_read: forall g s ha hb la lb
    (RCR: mm_refconfrel g s (ha, Nothing) (hb, Nothing))
    (REF: snd s la lb),
  exists va vb,
    <<RDa: pmread ha la va>> /\
    <<RDb: pmread hb lb vb>> /\
    <<VCL: vclos mm_gw s (g s) va vb>>.
Proof.
  i; rrdes RCR; clarify; eauto.
Qed.

Lemma mm_confrel_read: forall (lw: world_loca mm_gw) g s ha hb la lb
    (RH: confrel (wlift lw) g s (ha, Nothing) (hb, Nothing))
    (VLDa: pmvalid ha)
    (VLDb: pmvalid hb)
    (DB: snd (fst s) la lb),
  exists va vb,
    <<RDa: pmread ha la va>> /\
    <<RDb: pmread hb lb vb>> /\
    <<VCL: vclos (wlift lw) s (g s) va vb>>.
Proof.
  i; dup RH; rrdes RH.
  hexploit mm_confrel_shape; eauto; []; i; des.
  rrdes WHX; dup WHY0; rrdes WHY0; subst.
  hexploit mm_refconfrel_read; eauto; []; i; des.
  hexploit lwconf_extra; [exact WHY|]; intros [XTRa XTRb].
  inversion XTRa; inversion XTRb; clear XTRa XTRb; subst.
  rrdes RDa RDb; rrdes WHX0 CPa CPb; s in HPA; s in HPB.
  subst; noemp; clarify.        

  hexploit mm_refconfrel_read; eauto; []; i; des; esplits.
  - eapply pmread_plus; eauto; rewrite pmplus_swap; eauto.
  - eapply pmread_plus; eauto; rewrite pmplus_swap; eauto.
  - destruct s; eapply vclos_lift; eauto.
Qed.


(** Convenvience lemmas for reasoning about [IEwrite]. *)

Lemma mm_refconfrel_write: forall g s la lb va vb ha hb
    (RCR: mm_refconfrel g s ({la |-> va} ## ha, Nothing)
                            ({lb |-> vb} ## hb, Nothing))
    (RL: snd s la lb)
    va' vb' (VCL: vclos mm_gw s (g s) va' vb'),
  mm_refconfrel g s ({la |-> va'} ## ha, Nothing)
                    ({lb |-> vb'} ## hb, Nothing).
Proof.
  i; rrdes RCR; clarify.
  rr; esplits; eauto using pmvalid_plus_cell.
  - i; eapply pmread_plusD in ALC; des.
    + eapply pmread_cellD in ALC; des; clarify.
      eauto using pmread_plus, pmread_cell.
    + eapply DOMa; eauto using pmread_plus, pmplus_comm.
  - i; eapply pmread_plusD in ALC; des.
    + eapply pmread_cellD in ALC; des; clarify.
      eauto using pmread_plus, pmread_cell.
    + eapply DOMb; eauto using pmread_plus, pmplus_comm.
  - i; dup REG; apply RELV in REG; des.
    apply pmread_plusD in RDa; apply pmread_plusD in RDb; des.
    + apply pmread_cellD in RDa; apply pmread_cellD in RDb; des; subst.
      destruct s; esplits; last (exact VCL)
      ; eauto using pmread_plus, pmread_cell, pmvalid_plus_cell.
    + exfalso.
      apply pmread_cellD in RDb; des; subst.
      hexploit (refst_bij REG0 RL); i; des.
      specialize (X0 eq_refl). subst.
      rrdes RDa; subst.
      apply RELV in RL; des.
      rewrite pmplus_assoc in VLDa.
      apply pmvalid_plus_l in VLDa.
      apply pmvalid_cell_cell in VLDa; vauto.
    + exfalso.
      apply pmread_cellD in RDa; des; subst.
      hexploit (refst_bij REG0 RL); i; des.
      specialize (X eq_refl). subst.
      rrdes RDb; subst.
      apply RELV in RL; des.
      rewrite pmplus_assoc in VLDb.
      apply pmvalid_plus_l in VLDb.
      apply pmvalid_cell_cell in VLDb; vauto.
    + eauto 10 using pmread_plus, pmplus_comm, pmvalid_plus_cell.
Qed.

Lemma mm_confrel_write: forall (lw: world_loca mm_gw) g s la lb va vb ha hb
    (CR: confrel (wlift lw) g s ({la |-> va} ## ha, Nothing)
                                ({lb |-> vb} ## hb, Nothing))
    (RL: snd (fst s) la lb)
    va' vb' (VCL: vclos (wlift lw) s (g s) va' vb')
    (VLDa: pmvalid ({la |-> va} ## ha))
    (VLDb: pmvalid ({lb |-> vb} ## hb)),
  confrel (wlift lw) g s ({la |-> va'} ## ha, Nothing)
                         ({lb |-> vb'} ## hb, Nothing).
Proof.
  i; dup CR; rrdes CR.
  hexploit mm_confrel_shape; eauto; []; i; des.
  rrdes WHX; dup WHY0; rrdes WHY0; subst.
  hexploit mm_refconfrel_read; eauto; []; i; des.
  hexploit lwconf_extra; [exact WHY|]; intros [XTRa XTRb].
  inversion XTRa; inversion XTRb; clear XTRa XTRb; subst.
  rrdes RDa RDb; rrdes WHX0 CPa CPb; s in HPA; s in HPB.
  subst; noemp; clarify.

  rewrite -!pmplus_assoc in H0; rewrite -H0 in VLDa.
  apply pmcell_uniq in H0; [|by auto]; des; subst.
  rewrite -!pmplus_assoc in H1; rewrite -H1 in VLDb.
  apply pmcell_uniq in H1; [|by auto]; des; subst.

  rr; esplits; [|by eauto| |].
  - rr; esplits; [by eauto| | |].
    + clear VCL0; destruct s.
      inst_pairs; eapply mm_refconfrel_write; eauto using vclos_unlift.
    + eauto.
    + eauto.
  - noemp; s; rewrite -pmplus_assoc; auto.
  - noemp; s; rewrite -pmplus_assoc; auto.
Qed.


(** Convenvience lemmas for reasoning about [IEnew]. *)

Lemma confrel_newa: forall {lw: world_loca mm_gw} {g s s0 s1 s2 ha hb la va}
    (RH: confrel (wlift lw) g (s, s0, s1, s2) (ha, Nothing) (hb, Nothing))
    (VLD: pmvalid ({la |-> va} ## ha)),
  forall lb', ~ s1 la lb'.
Proof.
  i; rrdes RH.
  hexploit mm_confrel_shape; eauto; []; i; des.
  hexploit lwconf_extra; [exact WHY|]; intros [XTRa XTRb].
  inversion XTRa; inversion XTRb; clear XTRa XTRb.
  rrdes WHX WHY0 WHX0; rrdes CPa CPb.
  subst; noemp; simpl cplus in *; clarify.
  apply RELV in FH; des; rrdes RDa; subst.
  rewrite -pmplus_assoc pmplus_assoc in VLD; decomp_pmvalid VLD.
  apply pmvalid_cell_cell in VLD4; vauto.
Qed.

Lemma confrel_newb: forall {lw: world_loca mm_gw} {g s s0 s1 s2 ha hb lb vb}
    (RH: confrel (wlift lw) g (s, s0, s1, s2) (ha, Nothing) (hb, Nothing))
    (VLD: pmvalid ({lb |-> vb} ## hb)),
  forall la', ~ s1 la' lb.
Proof.
  i; rrdes RH.
  hexploit mm_confrel_shape; eauto; []; i; des.
  hexploit lwconf_extra; [exact WHY|]; intros [XTRa XTRb].
  inversion XTRa; inversion XTRb; clear XTRa XTRb.
  rrdes WHX WHY0 WHX0; rrdes CPa CPb.
  subst; noemp; simpl cplus in *; clarify.
  apply RELV in FH; des; rrdes RDb; subst.
  rewrite -pmplus_assoc pmplus_assoc in VLD; decomp_pmvalid VLD.
  apply pmvalid_cell_cell in VLD4; vauto.
Qed.

Lemma confrel_alloc: forall (lw: world_loca mm_gw) g s s0 s1 s2 va vb ha hb la lb
    (GK: gknow (wlift lw) g)
    (SIM: vclos (wlift lw) (s, s0, s1, s2) (g (s, s0, s1, s2)) va vb)
    (RH: confrel (wlift lw) g (s, s0, s1, s2) (ha, Nothing) (hb, Nothing))
    (VLDa: pmvalid ({la |-> va} ## ha))
    (VLDb: pmvalid ({lb |-> vb} ## hb))
    (FINa: pmfinite ({la |-> va} ## ha))
    (FINb: pmfinite ({lb |-> vb} ## hb))
    (s1': refst) (S1: refst_alloc s1 la lb = s1')
    (STA: stable lw),
  exists s2',
    extpub _ s2' s2 /\
    confrel (wlift lw) g (s, s0, s1', s2')
            ({la |-> va} ## ha, Nothing) ({lb |-> vb} ## hb, Nothing).
Proof.
  i; rrdes RH.
  hexploit mm_confrel_shape; eauto; i; des; clarify.
  destruct hay, hby; vauto; s in HPB; s in HPA; clarify.
  hexploit lwconf_extra; [exact WHY|]; intros [XTRa XTRb].
  inversion XTRa; inversion XTRb; clear XTRa XTRb; clarify.

  assert (CR: forall s2' (EXT: extpub _ s2' s2),
                confrel mm_gw (fun sg : state mm_gw => g (sg, s2')) (s, s0, s1')
                        ({la |-> va} ## ha', Nothing) ({lb |-> vb} ## hb', Nothing)).
  {
    i; rrdes WHX; destruct hax, hbx, hay, hby; vauto.
    destruct e, e1; vauto; destruct e0, e2; vauto.
    s in HPA; s in HPB; clarify.

    rr; esplits; eauto; last first.
    - inst_pairs; s; rewrite <-(pmplus_swap h0); eauto.
    - inst_pairs; s; rewrite <-(pmplus_swap h); eauto.
    - rrdes WHY0; clarify.
      assert (pmvalid ({la |-> va} ## ha')).
      { rewrite (pmplus_comm h) 2!pmplus_assoc in VLDa.
        decomp_pmvalid VLDa; eauto.
      }
      assert (pmvalid ({lb |-> vb} ## hb')).
      { rewrite (pmplus_comm h0) 2!pmplus_assoc in VLDb.
        decomp_pmvalid VLDb; eauto.
      }
      unfold refst_alloc in *.
      rr; esplits; eauto.
      + i; apply pmread_plusD in ALC; des.
        { s; rewrite <-S1; apply pmread_cellD in ALC; des; s; eauto. }
        { s; rewrite <-S1; hexploit DOMa; eauto; i; des; s; eauto. }
      + i; apply pmread_plusD in ALC; des.
        { s; rewrite <-S1; apply pmread_cellD in ALC; des; s; eauto. }
        { s; rewrite <-S1; hexploit DOMb; eauto; i; des; s; eauto. }
      + clean; rewrite <-S1.
        s; i; des; subst.
        { apply RELV in REG; des; esplits.
          - eapply pmread_plus; eauto.
            rewrite <-pmplus_comm; eauto.
          - eapply pmread_plus; eauto.
            rewrite <-pmplus_comm; eauto.
          - eapply vclos_mon; s; try rewrite <-S1; eauto.
            i; eapply (gk_mon GK); rr; s; esplits; eauto.
            i; rewrite -S1; eauto.
        }
        { esplits.
          - eapply pmread_plus; eauto using pmread_cell.
          - eapply pmread_plus; eauto using pmread_cell.
          - eapply vclos_unlift in SIM; eapply vclos_mon
            ; s; try rewrite <-S1; eauto.
            i; eapply (gk_mon GK); rr; s; esplits; eauto.
            i; rewrite -S1; eauto.
        }
  }

  exploit STA.
  { eauto. }
  { eauto. }
  { split; last first.
    - inst_pairs; eapply refst_alloc_ext; eauto.
    - s; auto.
  }
  { insts $ s0 s.
    set (dummy := (pmemp, Just (nil, IEunit)) : conf mid_lsi).
    rr; esplits; econs.
    - eauto.
    - insts $ dummy dummy; split; unfold dummy; s; eauto.
    - instantiate (2:=cemp); inst_pairs; s; noemp.
      unfold dummy; rewrite -pmplus_assoc; rr; esplits; eauto.
    - instantiate (2:=cemp); inst_pairs; s; noemp.
      unfold dummy; rewrite -pmplus_assoc; rr; esplits; eauto.
  }
  
  i; des; esplits; eauto.
  rr; esplits; clean; eauto
  ; try (by rewrite pmplus_assoc; inst_pairs; s; noemp; eauto).
Qed.



(** Tactic for destructing some basic world-related assumptions that show up
    in the compatibility proofs.
*)

Ltac mm_destruct XTRa XTRb CNF :=
  inversion XTRa; inversion XTRb; clear XTRa XTRb 
  ; inversion CNF as [RH CO MA MB]; clear CNF
  ; dup RH; destruct RH as (? & ? & ? & ? & WHX & WHY & ? & ?)
  ; apply mm_confrel_shape in WHX; des
  ; destruct MA as (? & ? & Eca & Ema & PVa & FINa)
  ; destruct MB as (? & ? & Ecb & Emb & PVb & FINb)
  ; unnw; subst; hexploit lwconf_extra; [exact WHY|]; intros [XTRa XTRb]
  ; inversion XTRa; inversion XTRb; clear XTRa XTRb; subst
  ; s in Eca; s in Ecb; clarify; noemp.


(** Convenience lemma for reasoning via esim_step.

    This basically says that when applying the esim_step case of
    esim_main, it is okay to assume that the given step on the target
    side was successful (not an error step), IF we can separately show
    that if the source can make a successful step, then so can the
    target (in other words: if the target makes an error step, then so
    does the source).

    TODO: To make the lemma more useful, we should add an assumption
    to the step case of mm_ok_esim_main saying that the source can
    take a successful step.
*)

Inductive mm_ok_esim_main (w: world mid_lsi mid_lsi) esimstep
  kia kib g si s cfa ma cfb mb : Prop :=
| mm_ok_esim_err
    mb' (STP: rtc (step _ etau) mb mb')
    (ER: mid_lsi.(error) mb')
| mm_ok_esim_step
    (NT: ~ mid_lsi.(halted) ma)
    (EST: forall evt ma' (STa: reduce_ok evt (fst ma) (fst (snd ma)) (snd (snd ma))
                                             (fst ma') (fst (snd ma')) (snd (snd ma'))),
          exists ea' csa' eb' csb' mb' mbt s' sync,
          <<STP: rtc (step _ etau) mb mb'>> /\
          <<ES: esimstep s' sync evt mb' mbt ea' eb'>> /\
          <<CNF: configure w g s' sync ea' eb' csa' csb' cfa cfb ma' mbt>> /\
          <<EXT: w.(ext) s' s>>)
| mm_ok_esim_ret s'
    mb' (STP: rtc (step _ etau) mb mb')
    va ea' (RETa: w.(cquerya) s' (ret va kia) ea')
    vb eb' (RETb: w.(cqueryb) s' (ret vb kib) eb')
    (RV: vclos w s' (g s') va vb)
    csa' csb' (CNF: configure w g s' true ea' eb' csa' csb' cfa cfb ma mb')
    (EXT: w.(ext) s' s) (EXTP: w.(extpub) s' si)
.

Lemma mm_esim_ok: forall w ha hb cfa cfb ea eb enva envb r kia kib g si s
    (NH: ~ exp_halted eb)
    (ST: forall evt hb' envb' eb'
                (STb: reduce_ok evt hb envb eb hb' envb' eb'),
         exists ha' enva' ea',
           reduce_ok evt ha enva ea ha' enva' ea')
    (SIM: forall evt hb' envb' eb'
                 (STb: reduce_ok evt hb envb eb hb' envb' eb'),
            mm_ok_esim_main w r kia kib g si s cfa (ha, (enva, ea)) cfb (hb, (envb, eb))),
     esim_main r kia kib g si s cfa (ha, (enva, ea)) cfb (hb, (envb, eb)).
Proof.
  i.
  destruct (LEM (forall evt hb' envb' eb',
                   ~ reduce_ok evt hb envb eb hb' envb' eb')).
  { econs; [econs; eapply IRerr|]; eauto. }
  do 4 (apply not_all_ex_not in f; des); apply NNPP in f.

  destruct (LEM (forall evt ha' enva' ea',
                   ~ reduce_ok evt ha enva ea ha' enva' ea')).
  - econs.
    + econs; eapply IRerr; [by eauto|].
      i; hexploit ST; eauto; i; des; eauto.
    + rr; i; rrdes FH; clarify; rrdes PV; vauto.
  - hexploit SIM; eauto; []; i; des.
    esplits; eauto; inv X; eauto.
    eapply esim_step; [by eauto|].
    i; inv ST0; eapply EST; eauto.
Qed.



(** * Program similarity.

    TODO: this should perhaps use a full world parameter, i.e. it may
    make also sense for other global worlds.
*)




Definition good_exp e := forall n, e <> IEnat n.

Definition clomem := list (var * exp).


Fixpoint lbls_upto (l: label) (G: list var) : list var :=
  match G with
    | inl (inl l') :: G' =>
      if eqdec l' l then nil else inl (inl l') :: lbls_upto l G'
    | _ :: G' =>
      lbls_upto l G'
    | nil => nil                
  end.

Definition drop_until_fun {X} {M: EqDec X} x x' (a : list X * bool) :=
  match a with
    | (acc, found) =>
        if found then 
          (acc, found)
        else 
          if eqdec x' x then
            (acc, true) 
          else
            (x'::acc, false) 
  end.

Definition drop_until' {X : Type} {M: EqDec X} (x: X) :=
  fold_right (drop_until_fun x) ([], false).

Lemma drop_until_cons {X : Type} {M: EqDec X} (x y: X) (xs : list X) :
  drop_until' y (x :: xs) = drop_until_fun y x (drop_until' y xs).
Proof. rewrite /drop_until'. s. auto. Qed.

Definition drop_until {X: Type} {M: EqDec X} (x: X) := fst <*> (drop_until' x).

Definition defbefore x G :=
  match x with
    | inl (inl l) => lbls_upto l G
    | _ => drop_until x G
  end.


Definition respects G (env: env) (phi: clomem) :=
  forall x e (IN: lookup x phi = Some e), exists xenv,
    <<IG: In x G>> /\
    <<IN: lookup x env = Some (IVclo xenv e)>> /\
    <<EN: forall y (IN: In y (defbefore x G)), lookup y xenv = lookup y env>>.

Definition fresh G x (phi: clomem) :=
  forall y e (IN: lookup y phi = Some e),
    x <> y /\ ~ In x (defbefore y G).

Definition mm_psim (lw: world_loca mm_gw) G ra phia phib ea eb : Prop :=
  <<GEa: good_exp ea>> /\
  <<GEb: good_exp eb>> /\
  exists i', forall g s si enva envb i kia kib
    (GK: gknow (wlift lw) g)
    (EXT: extpub _ s si)
    (RPa: respects G enva phia)
    (RPb: respects G envb phib)
    (RV: forall x (IN: In (inl x) G), exists va vb,
           <<INa: lookup (inl (rename_itvar ra x)) enva = Some va>> /\
           <<INb: lookup (inl x) envb = Some vb>> /\
           <<SIM: vclos (wlift lw) s (g s) va vb>>)
    (RK: forall x (IN: In (inr x) G), exists ka kb,
           <<INa: lookup (inr (rename_contvar ra x)) enva = Some ka>> /\
           <<INb: lookup (inr x) envb = Some kb>> /\
           <<SIM: ksim (wlift lw) (wfo_nat i) kia kib g si s s ka kb>>),
  esim (wlift lw) (wfo_nat (i' + i)) kia kib g si s true
        (pmemp, Just (enva, ea))
        (pmemp, Just (envb, eb)).


Definition mm_asim (lw: world_loca mm_gw) (G: list var) ra phia phib ea eb : Prop :=
  forall g s enva envb
    (GK: gknow (wlift lw) g)
    (RPa: respects G enva phia)
    (RPb: respects G envb phib)
    (RV: forall x (IN: In (inl x) G), exists va vb,
           <<INa: lookup (inl (rename_itvar ra x)) enva = Some va>> /\
           <<INb: lookup (inl x) envb = Some vb>> /\
           <<SIM: vclos (wlift lw) s (g s) va vb>>)
    vb (EV: eval envb eb = Some vb),
  exists va, eval enva ea = Some va /\
             vclos (wlift lw) s (g s) va vb.


Definition mm_ksim (lw: world_loca mm_gw) (G: list var) ra phia phib ea eb : Prop :=
  forall g s si i j kia kib enva envb bxb eb'
      (GK: gknow (wlift lw) g)
      (EXT: extpub _ s si)
      (RPa: respects G enva phia)
      (RPb: respects G envb phib)
      (RV: forall x (IN: In (inl x) G), exists va vb,
             <<INa: lookup (inl (rename_itvar ra x)) enva = Some va>> /\
             <<INb: lookup (inl x) envb = Some vb>> /\
             <<SIM: vclos (wlift lw) s (g s) va vb>>)
      (RK: forall x (IN: In (inr x) G), exists ka kb,
             <<INa: lookup (inr (rename_contvar ra x)) enva = Some ka>> /\
             <<INb: lookup (inr x) envb = Some kb>> /\
             <<SIM: ksim (wlift lw) (wfo_nat i) kia kib g si s s ka kb>>)
      (Eb: eb = IEcont bxb eb'),
    exists bxa ea',
      <<Ea: ea = IEcont bxa ea'>> /\
      <<SIM: ksim (wlift lw) j kia kib g si s s (IVclo enva ea) (IVclo envb eb)>>.



                  


Lemma foo': forall (X : Type) (M : EqDec X) (Y Z : Type) 
         (f : Z -> X) (g : Z -> Z) (xs : list X) (xysa xysb : list (X * Y))
         (R : Y -> Y -> Prop) (xa' xb': X) (ya' yb' : Y)
         x (IN: In (f x) (xb' :: xs))
         (RR: f x = xb' -> f (g x) = xa' /\ R ya' yb')
         (KA: f (g x) = xa' -> f x = xb')
         (RST: forall (x : Z) (IN: In (f x) xs) (NEb: f x <> xb') (NEa: f (g x) <> xa'),
                 exists ya yb : Y,
                 lookup (f (g x)) xysa = Some ya /\
                 lookup (f x) xysb = Some yb /\ R ya yb),
  exists ya yb : Y,
    lookup (f (g x)) ((xa', ya') :: xysa) = Some ya /\
    lookup (f x) ((xb', yb') :: xysb) = Some yb /\ R ya yb.
Proof.
  i; apply in_consD in IN; des.
  - subst; rewrite !list_assoc_cons_eq.
    specialize (RR eq_refl); des; rewrite RR. eauto.
  - rewrite !list_assoc_cons_neq; eauto.
Qed.

Definition extendable_t (G: list var) ren xa xb :=
  forall xb'
      (IN: In (inl (inr xb')) G)
      (E: rename_termvar ren xb' = inr xa),
    xb' = xb.
    

Definition extendable_c (G: list var) ren xa xb :=
  forall xb'
      (IN: In (inr xb') G)
      (E: rename_contvar ren xb' = xa),
    xb' = xb.
    

Lemma bla: forall G ra bxa bxb bya byb
    (EX: extendable_t G ra bxa bxb)
    (EE: bya = bxa -> byb = bxb),
  extendable_t (inl (inr byb) :: G) (ra)[inr bya/byb] bxa bxb.
Proof.
  ii; destruct (LEM (xb' = byb)).
  - subst; s in E; tmext_simplH E; clarify; eauto.
  - s in E; rewrite tmext_neq in E; eauto.
    exploit EX; eauto; []; i; des; subst; eauto.
    apply in_consD in IN; des; vauto.
Qed.

Definition nolbl (x: var) :=
  forall l, x <> inl (inl l).   

Lemma nolbl_t: forall x,
  nolbl (inl (inr x)).
Proof.
 vauto.
Qed.
Hint Resolve nolbl_t.

Lemma nolbl_c: forall x,
  nolbl (inr x).
Proof.
  vauto.
Qed.
Hint Resolve nolbl_c.

Lemma drop_until_in {X} {M: EqDec X} x xs :
  In x xs ->
  snd (drop_until' x xs) = true.
Proof.
  induction xs; auto. ii; apply in_consD in H; des; clarify.
  + rewrite drop_until_cons. destruct (drop_until' a xs); destruct b; s; auto; destruct eqdec; auto.
  + specialize (IHxs H). rewrite drop_until_cons.
    destruct (drop_until' x xs); destruct b; auto.
Qed.    

Lemma drop_until_step {X} {M: EqDec X} x a xs :
  In x xs ->
  (drop_until' x (a :: xs)) = (drop_until' x xs).
Proof.
  revert x a. induction xs; ii; s in H; first case H; clarify.
  apply in_consD in H; des; clarify.
  + rewrite !drop_until_cons; s.
    destruct drop_until'; s; rewrite !eqdec_true. destruct b; auto.
  + specialize (IHxs x a H).
    assert (H2 := drop_until_in _ _ H).
    revert H2 IHxs. rewrite !drop_until_cons. destruct drop_until'; s; intro. rewrite H2. auto.
Qed.

    
Lemma defbefore_cons: forall x y G
    (IN: In x G)
    (NL: nolbl y),
  defbefore x G = defbefore x (y::G).
Proof.
  ii. destruct y as [[ly|ty]|ky], x as [[lx|tx]|kx]; try (by exfalso; eapply NL; eauto); eauto;
  rewrite {2}/defbefore; s drop_until; rewrite /drop_until; unfold compose; rewrite (drop_until_step _ _ _ IN);
  auto.
Qed.

Lemma drop_until_sublist {X} `{M: EqDec X} x (xs: list X):
  sublist (drop_until x xs) xs.
Proof. 
  induction xs; s; eauto using sublist_nil'. revert IHxs. 
  rewrite /drop_until. unfold compose. rewrite drop_until_cons.
  destruct drop_until'; destruct b. eauto.
  s; destruct eqdec; s; eauto.
Qed. 

Lemma drop_until_purge x xs :
  drop_until' (inl x) (purge xs) = (purge (fst (drop_until' (inl x) xs)), snd (drop_until' (inl x) xs)).
Proof.
  induction xs; auto.
  destruct a as [[la|ta]|ka]; rewrite !drop_until_cons; s;
  destruct x as [lx|tx]; s; eauto;
  rewrite IHxs; s;
  destruct (drop_until' (inl _) xs); s; destruct b; s; auto;
  destruct eq_nat_dec; clarify; s; auto.
Qed.

Lemma defbefore_purge x xs :
  defbefore (inl x) (purge xs) = purge (defbefore (inl x) xs).
Proof.
  ii. induction xs; auto; destruct x as [lx|tx]; eauto.
  - s. destruct a as [[la|ta]|ka]; s; eauto.
    destruct eq_nat_dec; eauto.
    s. s in IHxs. by rewrite IHxs.
  - rewrite /defbefore /drop_until. s.
    by rewrite (drop_until_purge (inr tx) (a::xs)).
Qed.

Lemma drop_until_not_in {X} {M: EqDec X} x xs :
  ~ In x xs ->
  snd (drop_until' x xs) = false.
Proof.
  induction xs; auto; ii; try apply nin_consD in H; des; clarify.
  + specialize (IHxs H0); rewrite drop_until_cons
    ; destruct (drop_until' x xs); destruct b; auto; []
    ; s; rewrite (eqdec_false H); vauto.
Qed.

Lemma drop_until_not_step {X} {M: EqDec X} x xs :
  ~ In x xs ->
  fst (drop_until' x (x :: xs)) = xs.
Proof.
  revert x. induction xs; ii; try apply nin_consD in H; des.
  + rewrite /drop_until'; s; rewrite eqdec_true; auto.
  + specialize (IHxs x H0). assert (H2 := drop_until_not_in _ _ H0).
    revert H2 IHxs.
    rewrite !drop_until_cons.
    destruct (drop_until' x xs); s
    ; intro; rewrite H2; rewrite eqdec_true; rewrite (eqdec_false H)
    ; i; clarify; s; rewrite eqdec_true; auto.
Qed.

Lemma drop_until_not_step' {X} {M: EqDec X} x a xs:
    ~ In x xs ->
    a <> x ->
    fst (drop_until' x (a::xs)) = a::xs.
Proof.
  revert x. induction xs; ii; try apply nin_consD in H; des.
  + rewrite /drop_until'; s; rewrite (eqdec_false H0); auto.
  + specialize (IHxs x H1 H0); assert (H2 := drop_until_not_in _ _ H1)
    ; revert H2 IHxs; rewrite !drop_until_cons; destruct (drop_until' x xs); s
    ; intro; rewrite H2; rewrite (eqdec_false H0); rewrite (eqdec_false H)
    ; i; clarify; s; rewrite (eqdec_false H0); vauto
    ; []; s in IHxs; subst; vauto.
Qed. 

Lemma defbefore_des: forall x G
    (NIN: ~ In x G)
    (NL: nolbl x),
  defbefore x (x::G) = G.
Proof.
  ii. destruct x as [[lx|tx]|kx]
  ; try (by exfalso; eapply NL; eauto); eauto
  ; rewrite /defbefore; rewrite /drop_until; unfold compose
  ; rewrite (drop_until_not_step _ _ NIN); auto.
Qed.

Lemma defbefore_des': forall x G
    (NIN: ~ In x G)
    (NL: nolbl x),
  defbefore x G = G.
Proof.
  ii. destruct x as [[lx|tx]|kx]
  ; try (by exfalso; eapply NL; eauto); eauto
  ; rewrite /defbefore; rewrite /drop_until; unfold compose; induction G
  ; vauto
  ; apply nin_consD in NIN; des; specialize (IHG NIN0)
  ; rewrite (drop_until_not_step' G NIN0 NIN); vauto.
Qed.

Lemma lbls_upto_sublist x (xs: list var):
  sublist (lbls_upto x xs) xs.
Proof. induction xs; eauto. s; destruct a as [[l|t]|k]; eauto. case eq_nat_dec; eauto using sublist_nil'. Qed.

Lemma lbls_upto_sublist' x (xs : list var):
  sublist (lbls_upto x (purge xs)) (lbls_upto x xs).
Proof.
  induction xs; auto.
  destruct a as [[l|t]|k]; s; first destruct eq_nat_dec; eauto.
Qed.
  

Lemma defbefore_sublist x (xs : list var):
  sublist (defbefore x xs) xs.
Proof. destruct x as [[l|t]|k]; s.
  + eapply sublist_trans. by eapply lbls_upto_sublist. eauto.
  + eapply drop_until_sublist.
  + eapply drop_until_sublist.
Qed.

Lemma bvars_nolbl e x :
  In x (bvars e) -> nolbl x.
Proof.
  induction e; s; ii;
  try (destruct bx; s in H; des);
  des; subst; clarify;
  try (eapply in_app_or in H);
  des; subst; clarify;
  try (by eapply IHe);
  try (by eapply IHe1);
  try (by eapply IHe2).
Qed.

Lemma purge_lbl_nolbl : forall G x,
  nolbl x ->
  In x G -> 
  In x (purge_lbl G).
Proof. ii. destruct x as [[]|]; try (by exfalso; eapply H); eauto. Qed.

Lemma respects_cons: forall G env phi x v
    (RE: respects G env phi)
    (NL: nolbl x)
    (FR: fresh G x phi),
  respects (x :: G) ((x, v) :: env) phi.
Proof.
  ii; exploit RE; eauto; []; i; des; subst.
  apply FR in IN; []; des; esplits.
  - s; eauto.
  - rewrite list_assoc_cons_neq; vauto.
  - rewrite -defbefore_cons; auto.
    i; destruct (LEM (x = y)); vauto.
    rewrite list_assoc_cons_neq; eauto.
Qed.

Lemma mm_compat_input: forall lw G ra bx phia phib ea eb
    (EX: extendable_t G ra bx bx)
    (FRa: fresh G (inl (inr bx)) phia)
    (FRb: fresh G (inl (inr bx)) phib)
    (IH: mm_psim lw (inl (inr bx) :: G) ra[inr bx/bx] phia phib ea eb),
  mm_psim lw G ra phia phib (IEinput bx ea) (IEinput bx eb).
Proof.
  i; rr; splits; vauto; []; exists 0; i; s.
  rrdes IH; pfold; econs; i.
  mm_destruct XTRa XTRb CNF.
  apply mm_esim_ok; [by eauto| |].
  - i; invc STb; eauto.
  - esplits; eauto; eapply mm_ok_esim_step; eauto; i.
    destruct ma' as [? []]; s in STa; invc STa.
    do 7 eexists; exists false; splits; first (by eauto)
    ; last (by s; eauto); first 1 last.
    + econs; eauto; noemp; inst_pairs; s; noemp; rr; eauto 20.
    + left; econs; eauto.
      * left; rr; eauto.
      * { eapply gres_incl_id; left.
          rewrite 2!mid_csplit.
          eapply pesim_release; s; eauto.
          eapply IH1; [by eauto|by s; eauto| | | |]; eauto using respects_cons.
          - i; eapply foo'; eauto.
            { i; clarify; s; tmext_simpl; split; eauto.
              econs; eapply vclgen_nat; s; eauto.
            }
            { i; clarify.
              destruct (LEM (x2 = inr bx)); vauto.
              rewrite ren_ext_it_neq in H; [|by ii; vauto].
              destruct x2; vauto.
              exploit (EX t); s in H; i; des; vauto.
              apply in_consD in IN; des; vauto.
            } 
            clear IN; i; rewrite ren_ext_it_neq; [by eauto|by ii; vauto].
          - i; apply in_consD in IN; des; clarify.
            rewrite !list_assoc_cons_neq; auto; ii; clarify.
            rewrite ren_ext_c_neq; auto.
        }
Qed.


Lemma extendable_t_same: forall G ren x
    (E: forall y (IN: In y G) (E: rename_var ren y = inl (inr x)), rename_var ren y = y),
  extendable_t G ren x x.
Proof.
  s; ii; clarify; hexploit (E (inl (inr xb'))); [|s; rewrite E0|]; eauto.
  s; i; clarify; rewrite H in E0; clarify.
Qed.


Lemma mm_compat_output: forall lw G ra xa xb phia phib ea eb
    (WF: In (inl xb) G)
    (E: xa = rename_itvar ra xb)
    (IH: mm_psim lw G ra phia phib ea eb),
  mm_psim lw G ra phia phib (IEoutput xa ea) (IEoutput xb eb).
Proof.
  i; rr; splits; vauto; []; exists 0; i; s.
  apply RV in WF.
  rrdes IH; pfold; econs; i.
  mm_destruct XTRa XTRb CNF.
  eapply mm_esim_ok; [by eauto| |].
  - i; invc STb; clarify.
    eapply mm_vclosDb in WF1; subst; eauto; des; clarify; eauto.
  - esplits; eauto; eapply mm_ok_esim_step; eauto; i.
    destruct ma' as [? []]; s in STa; invc STa.
    do 7 eexists; exists false; splits; first (by eauto)
    ; last (by s; eauto); first 1 last.
    + econs; eauto; noemp; inst_pairs; s; noemp; rr; eauto 10.
    + left; econs; eauto.
      * invc STb; rewrite WF in GET; clarify.
        left; eapply mm_vclosDb in WF1; eauto; des; clarify.
        eapply IRok; eauto.
      * eapply gres_incl_id; left.
        rewrite 2!mid_csplit.
        eapply pesim_release; s; eauto; eapply IH1; eauto.
Qed.

Lemma ren_itvar_id: forall x,
  rename_itvar ren_id x = x.
Proof.
    by destruct x as [[]|].
Qed.


Lemma mm_compat_app: forall lw G ra phia phib xfa xfb xva xvb xka xkb
    (INf: In (inl xfb) G)
    (INv: In (inl xvb) G)
    (INk: In (inr xkb) G)
    (Ef: xfa = rename_itvar ra xfb)
    (Ev: xva = rename_itvar ra xvb)
    (Ek: xka = rename_contvar ra xkb),
  mm_psim lw G ra phia phib (IEapp xfa xva xka) (IEapp xfb xvb xkb).
Proof.
  i; rr; splits; vauto; []; exists 0; i; s.
  apply RV in INf; apply RV in INv; apply RK in INk; des.
  pfold; econs; i.
  mm_destruct XTRa XTRb CNF.
  eapply mm_esim_ok; [by eauto| |].
  - i; invc STb.
    clarify.
    eapply mm_vclosDb in INf1; des; clarify.
    esplits; eauto using IRapp.
  - esplits; eauto; eapply mm_ok_esim_step; eauto; i.
    destruct ma' as [? []]; s in STa; invc STa STb.
    do 7 eexists; exists true; esplits; first (by eauto); first 1 last; eauto.
    + econs; first (by eauto); first 1 last.
      * inst_pairs; s; noemp; rr; instantiate (2 := pmemp); s; noemp; eauto 10.
      * inst_pairs; s; noemp; rr; instantiate (5 := pmemp); s; noemp; eauto 10.
      * clear CO CO0; s; eauto 10.
    + clarify.
      rewrite INv in GETV; rewrite INf in GET; clarify.
      right; econs; eauto.
      * econs. 
        hexploit (IRapp ((hb' ## x3) ## x0) envb); eauto.
      * { econs 1.
          - rr; eauto.
          - rr; eauto.
          - apply mm_vclosDb in INf1; des.
            econs 1; rr; eauto.
          - eauto.
          - eapply ksim_mon; eauto using gres_incl_id.
        }
Qed.


Lemma mm_compat_read: forall lw G ra phia phib xa xb bx ea eb
    (WF: In (inl xb) G)
    (E: xa = rename_itvar ra xb)
    (EX: extendable_t G ra bx bx)
    (FRa: fresh G (inl (inr bx)) phia)
    (FRb: fresh G (inl (inr bx)) phib)
    (IH: mm_psim lw (inl (inr bx) :: G) ra[inr bx/bx] phia phib ea eb),
  mm_psim lw G ra phia phib (IEread xa bx ea) (IEread xb bx eb).
Proof.
  i; rr; splits; vauto; []; exists 0; i; s.
  apply RV in WF; des.
  rrdes IH; pfold; econs; i.
  mm_destruct XTRa XTRb CNF.
  eapply mm_esim_ok; [by eauto| |].
  - i; invc STb.
    clarify.
    eapply mm_vclosDb in WF1; des; clarify.
    rrdes QR; clarify.
    dup PVa; apply pmvalid_plusD in PVa; des.
    dup PVb; apply pmvalid_plusD in PVb; des.
simpl cplus in RH0.
    eapply mm_confrel_read in RH0; eauto; []; des.
    esplits; econs; eauto using pmread_plus.
  - esplits; eauto; eapply mm_ok_esim_step; eauto; i.
    destruct ma' as [? []]; s in STa; invc STa STb.
    rewrite WF in GET; clarify.
    dup WF1; eapply mm_vclosDb in WF1; []; des; clarify.
    rrdes QR; clarify.
    dup PVa; apply pmvalid_plusD in PVa; des.
    dup PVb; apply pmvalid_plusD in PVb; des.
simpl cplus in RH0.
    dup RH0; eapply mm_confrel_read in RH0; eauto; []; des.
    eapply pmread_plus in RDa; eauto.
    eapply pmread_plus in RDb; eauto.
    hexploit (pmread_uniq RDa LOAD); []; i; inst; subst.
    hexploit (pmread_uniq RDb LOAD0); []; i; inst; subst.
    do 7 eexists; exists false; splits; first (by eauto); last (by s; eauto)
    ; last (econs; [by eauto|by eauto| |]); last first; first 1 last.
    + inst_pairs; s; noemp; rr; esplits; eauto.
    + left; econs; eauto.
      * left; inst_pairs; eapply IRok; econs; eauto.
      * eapply gres_incl_id; left.
        rewrite mid_csplit; instantiate (1:=_#_).
        clear CO0 CO.
        eapply pesim_release; eauto; try (by rr; eauto).
        { hexploit IH1; [M| |M|M| | |M]; (Mdo by eauto using respects_cons).
        - eapply extpub_trans; [|by eauto]; s; eauto.
        - i; eapply foo'; eauto.
          { i; clarify; s; tmext_simpl; eauto. }
          { i; clarify.
            destruct (LEM (x2 = inr bx)); vauto.
            rewrite ren_ext_it_neq in H; [|by ii; vauto].
            destruct x2; vauto; exploit (EX t); s in H; i; vauto.
            apply in_consD in IN; des; vauto.
          } 
          clear IN; i; rewrite ren_ext_it_neq; [by eauto|by ii; vauto].
        - i; apply in_consD in IN; des; clarify.
          rewrite !list_assoc_cons_neq; auto; ii; clarify.
          rewrite ren_ext_c_neq; auto.
        }
    + s; noemp; rr; esplits; eauto 10.
Qed.


Lemma mm_compat_write: forall lw G ra phia phib xa xb ya yb ea eb
    (WFx: In (inl xb) G)
    (WFy: In (inl yb) G)
    (Ex: xa = rename_itvar ra xb)
    (Ey: ya = rename_itvar ra yb)
    (IH: mm_psim lw G ra phia phib ea eb),
  mm_psim lw G ra phia phib (IEwrite xa ya ea) (IEwrite xb yb eb).
Proof.
  i; rr; splits; vauto; []; exists 0; i; s.
  apply RV in WFx; apply RV in WFy; des.
  rrdes IH; pfold; econs; i; invc CNF XTRa XTRb.
  dup RH; rrdes RH.
  hexploit mm_confrel_shape; eauto; []; i; des.
  rrdes WHX; dup WHY0; rrdes WHY0; subst.
  hexploit lwconf_extra; [exact WHY|]; intros [XTRa XTRb].
  inversion XTRa; inversion XTRb; clear XTRa XTRb; subst.
  rrdes MA MB; subst; s in Ec; s in Ec0; clarify; noemp.
  eapply mm_esim_ok; [by eauto| |].
  - i; invc STb.
    clarify.
    eapply mm_vclosDb in WFx1; des; clarify.
    dup PV; apply pmvalid_plusD in PV; des.
    dup PV0; apply pmvalid_plusD in PV0; des.
    simpl cplus in RH0; rrdes QR; clarify.
    eapply mm_confrel_read in REG; eauto; des.
    eapply pmread_plus in RDa; eauto; rrdes RDa; rewrite RD.
    esplits; econs; eauto.
  - esplits; eauto; eapply mm_ok_esim_step; eauto; i.
    destruct ma' as [? []]; s in STa; invc STa STb.
    clarify.
    rewrite WFy in GET0; rewrite WFx in GET; clarify.
    dup WFx1; eapply mm_vclosDb in WFx1; eauto; []; des; clarify.
    dup PV; apply pmvalid_plusD in PV; des.
    dup PV0; apply pmvalid_plusD in PV0; des.
    simpl cplus in RH0; rrdes QR; clarify.
    dup REG; eapply mm_confrel_read in REG; eauto; des.
    dup RDa; rrdes RDa; subst; rewrite RD in H1, PV1.
    rewrite -pmplus_assoc in H1, PV1; rewrite -H1 in PV1.
    apply pmcell_uniq in H1; [|by auto]; des; subst.
    dup RDb; rrdes RDb; subst; rewrite RD0 in H2, PV3.
    rewrite -pmplus_assoc in H2, PV3; rewrite -H2 in PV3.
    apply pmcell_uniq in H2; [|by auto]; des; subst.
    rrdes WHX0 CPa CPb; clarify; noemp; clarify.
    do 7 eexists; exists false; splits; first (by eauto); last (by s; eauto)
    ; last (econs; [by eauto|by eauto| |]); last first; first 1 last.
    + rewrite pmplus_assoc; inst_pairs; s; noemp; rr; esplits; eauto.
      * rewrite <-pmplus_assoc; eauto using pmvalid_plus_cell.
      * rewrite RD -pmplus_assoc in FIN.
        rewrite <-pmplus_assoc; eauto using pmfinite_plus_cell.
    + left; econs; eauto.
      * left; inst_pairs; eapply IRok; econs; eauto.
      * eapply gres_incl_id; left.
        rewrite mid_csplit; instantiate (1:=_#_).
        clear CO CO0.
        eapply pesim_release; eauto; try (by rr; eauto).
        { hexploit IH1; eauto using extpub_trans. }
        { instantiate (1:= ({lb |-> v0} ## h'0, Nothing)).
          rewrite RD RD0 in RH0, VLD, VLD0.
          eapply mm_confrel_write; eauto.
        }
    + s; noemp.
      rr; rewrite <-pmplus_assoc; esplits; eauto using pmvalid_plus_cell.
      rewrite RD0 -pmplus_assoc in FIN0; eauto using pmfinite_plus_cell.
Qed.


Lemma mm_compat_new: forall lw G ra phia phib xa xb bx ea eb
    (WF: In (inl xb) G)
    (E: xa = rename_itvar ra xb)
    (FRa: fresh G (inl (inr bx)) phia)
    (FRb: fresh G (inl (inr bx)) phib)
    (EX: extendable_t G ra bx bx)
    (IH: mm_psim lw (inl (inr bx) :: G) ra[inr bx/bx] phia phib ea eb)
    (STA: stable lw),
  mm_psim lw G ra phia phib (IEnew xa bx ea) (IEnew xb bx eb).
Proof.
  i; rr; splits; vauto; []; exists 0; i; s.
  apply RV in WF; des.
  rrdes IH; pfold; econs; i.
  mm_destruct XTRa XTRb CNF.
  eapply mm_esim_ok; [by eauto| |].
  - i; invc STb.
    clarify.
    hexploit (@pmfinite_alloc _ _ _ ((ha' ## x1) ## x)); eauto. 
    i; des; esplits; econs; eauto.
  - esplits; eauto; eapply mm_ok_esim_step; eauto; i.
    destruct ma' as [? []]; s in STa; invc STa.
    hexploit (@pmfinite_alloc _ _ _ ((hb' ## x3) ## x0)); eauto.
    i; des; destruct s as [[[]]].
    dup H; rewrite pmplus_assoc in H0; apply pmvalid_plus_l in H0.
    dup PV; rewrite pmplus_assoc in PV0; apply pmvalid_plus_l in PV0.
    simpl cplus in RH0.
    rewrite WF in GET; clarify.
    hexploit confrel_alloc.
    { eauto. }
    { eapply vclos_mon; eauto. }
    { eapply confrel_mon; eauto. }
    { eauto. }
    { eauto. }
    { split_pmfinite FINa PVa; eauto. }
    { split_pmfinite FINb PVb; eauto. }
    { instantiate (1 := refst_allocP s1 loc x2 (confrel_newa RH0 PV0) (confrel_newb RH0 H0)).
      eauto.
    }
    { eauto. }
    i; des.
    do 6 econs
    ; exists (s, s0, refst_allocP s1 loc x2 (confrel_newa RH0 PV0) (confrel_newb RH0 H0), s2')
    ; exists false.
    splits; first (by eauto); last (by s; unfold refst_alloc; eauto)
    ; last (econs; [by eauto|by eauto| |]); last first; first 1 last.
    + inst_pairs; s; noemp; rewrite pmplus_assoc.
      rr; esplits; eauto; inst; rewrite <-pmplus_assoc; eauto.
    + left; econs; auto; [left; inst_pairs; econs 1|].
      eapply IRnew; eauto; econs; eauto.
      eapply gres_incl_id; left.
      rewrite mid_csplit; instantiate (1:=_#_).
      clear CO CO0.
      eapply pesim_release; eauto; try (by rr; eauto).
      { hexploit IH1; [M| |M|M| | |M]; (Mdo by eauto using respects_cons).
        - eapply extpub_trans; [|by eauto]; s; unfold refst_alloc; eauto.
        - i; eapply foo'; eauto.
          { i; clarify; s; tmext_simpl; split; eauto.
            econs; eapply vclgen_ref.
            rr; esplits; s; unfold refst_alloc; eauto.
          }
          { i; clarify.
            destruct (LEM (x4 = inr bx)); vauto.
            rewrite ren_ext_it_neq in H3; [|by ii; vauto].
            destruct x4; vauto.
            exploit (EX t); s in H; i; des; vauto.
            apply in_consD in IN; des; vauto.
          } 
          clear IN; i; rewrite ren_ext_it_neq; [|by ii; vauto].
          apply RV in IN; des; esplits; eauto.
          eapply vclos_mon; s; unfold refst_alloc; eauto.
          i; eapply gk_mon; s; unfold refst_alloc; eauto.
        - i; apply in_consD in IN; des; clarify.
          rewrite ren_ext_c_neq !list_assoc_cons_neq; try by auto.
          apply RK in IN; des; esplits; eauto;
          eapply ksim_mon; s; unfold refst_alloc; eauto.
      }
    + s; unfold mid_val; noemp.
      rr; rewrite <-pmplus_assoc; esplits; eauto.
Qed.


Lemma mm_compat_ret: forall lw G ra phia phib xka xkb xva xvb
    (INk: In (inr xkb) G)
    (INv: In (inl xvb) G)
    (Ev: xva = rename_itvar ra xvb)
    (Ek: xka = rename_contvar ra xkb),
  mm_psim lw G ra phia phib (IEret xka xva) (IEret xkb xvb).
Proof.
  i; rr; splits; vauto; []; exists 0; i; s.
  apply RK in INk; apply RV in INv; des.
  inv INk1; eapply KR; s; eauto.
Qed.


Lemma mm_compat_ifnz: forall lw G ra phia phib xa xb ea ea' eb eb'
    (WF: In (inl xb) G)
    (E: xa = rename_itvar ra xb)
    (IH: mm_psim lw G ra phia phib ea eb)
    (IH': mm_psim lw G ra phia phib ea' eb'),
  mm_psim lw G ra phia phib (IEifnz xa ea ea') (IEifnz xb eb eb').
Proof.
  i; rr; splits; vauto; []; exists 0; i; s.
  apply RV in WF; des.
  rrdes IH IH'; pfold; econs; i.
  mm_destruct XTRa XTRb CNF.
  eapply mm_esim_ok; [by eauto| |].
  - i; inv STb
    ; clarify
    ; eapply mm_vclosDb in WF1; eauto; des; clarify; eauto.
  - esplits; eauto; eapply mm_ok_esim_step; eauto; i.
    destruct ma' as [? []]; s in STa; invc STb.
    { clarify.
      apply mm_vclosDb in WF1; []; subst.
      invc STa; [|by rewrite WF in GET; clarify].
      do 7 eexists; exists false; splits; first (by eauto); last (by s; eauto)
      ; first 1 last.
      + econs; eauto; noemp; inst_pairs; s; noemp; rr; eauto 10.
      + left; econs; eauto.
        * left; eapply IRok; eauto.
        * eapply gres_incl_id; left.
          rewrite 2!mid_csplit.
          eapply pesim_release; eauto; try (by rr; eauto).
          eapply IH'1; eauto.
    }
    { clarify.
      apply mm_vclosDb in WF1; []; subst.
      invc STa; [by rewrite WF in GET; clarify|].
      do 7 eexists; exists false; splits; first (by eauto); last (by s; eauto)
      ; first 1 last.
      + econs; eauto; noemp; inst_pairs; s; noemp; rr; eauto 10.
      + left; econs; eauto.
        * left; eapply IRok; eauto.
        * eapply gres_incl_id; left.
          rewrite 2!mid_csplit.
          eapply pesim_release; eauto; try (by rr; eauto).
          eapply IH1; eauto.
    }
Qed.


Lemma mm_compat_case: forall lw G ra phia phib xa xb bx ea ea' eb eb'
    (WF: In (inl xb) G)
    (E: xa = rename_itvar ra xb)
    (FRa: fresh G (inl (inr bx)) phia)
    (FRb: fresh G (inl (inr bx)) phib)
    (EX: extendable_t G ra bx bx)
    (IH: mm_psim lw (inl (inr bx) :: G) ra[inr bx/bx] phia phib ea eb)
    (IH': mm_psim lw (inl (inr bx) :: G) ra[inr bx/bx] phia phib ea' eb'),
  mm_psim lw G ra phia phib (IEcase xa bx ea ea') (IEcase xb bx eb eb').
Proof.
  i; rr; splits; vauto; []; exists 0; i; s.
  apply RV in WF; des.
  rrdes IH IH'; pfold; econs; i.
  mm_destruct XTRa XTRb CNF.
  eapply mm_esim_ok; [by eauto| |].
  - i; inv STb
    ; clarify
    ; eapply mm_vclosDb in WF1; eauto; des; clarify
    ; eauto using IRcasel, IRcaser.
  - esplits; eauto; eapply mm_ok_esim_step; eauto; i.
    destruct ma' as [? []]; s in STa; invc STb.
    { clarify.
      apply mm_vclosDb in WF1; []; des; subst.
      invc STa; [|by rewrite WF in GET0; clarify].
      do 7 eexists; exists false; splits; first (by eauto); last (by s; eauto)
      ; first 1 last.
      + econs; eauto; noemp; inst_pairs; s; noemp; rr; eauto 10.
      + left; econs; eauto.
        * left; eapply IRok; eauto using IRcasel.
        * eapply gres_incl_id; left.
          rewrite 2!mid_csplit.
          eapply pesim_release; eauto; try (by rr; eauto).
          rewrite WF in GET0; clarify.
          eapply IH1; eauto using respects_cons; [|by i; eapply foo'; eauto; vauto].
          i; eapply foo'; eauto; vauto.
          { i; clarify; s; tmext_simpl; eauto. }
          { i; clarify.
            destruct (LEM (x2 = inr bx)); vauto.
            rewrite ren_ext_it_neq in H; [|by ii; vauto].
            destruct x2; vauto; exploit (EX t); s in H; i; vauto.
            apply in_consD in IN; des; vauto.
          }
          clear IN; i; rewrite ren_ext_it_neq; [eauto|ii; vauto].
    }
    { clarify.
      apply mm_vclosDb in WF1; []; des; subst.
      invc STa; [by rewrite WF in GET0; clarify|].
      do 7 eexists; exists false; splits; first (by eauto); last (by s; eauto)
      ; first 1 last.
      + econs; eauto; noemp; inst_pairs; s; noemp; rr; eauto 10.
      + left; econs; eauto.
        * left; eapply IRok; eauto using IRcaser.
        * eapply gres_incl_id; left.
          rewrite 2!mid_csplit.
          eapply pesim_release; eauto; try (by rr; eauto).
          rewrite WF in GET0; clarify.
          eapply IH'1; eauto using respects_cons; [|by i; eapply foo'; eauto; vauto].
          i; eapply foo'; eauto; vauto. 
          { i; clarify; s; tmext_simpl; eauto. }
          { i; clarify.
            destruct (LEM (x2 = inr bx)); vauto.
            rewrite ren_ext_it_neq in H; [|by ii; vauto].
            destruct x2; vauto; exploit (EX t); s in H; i; vauto.
            apply in_consD in IN; des; vauto.
          }
          clear IN; i; rewrite ren_ext_it_neq; [eauto|ii; vauto].
    }
Qed.


Lemma mm_compat_inst: forall lw G ra phia phib xfa xfb xka xkb
    (INf: In (inl xfb) G)
    (INk: In (inr xkb) G)
    (Ef: xfa = rename_itvar ra xfb)
    (Ek: xka = rename_contvar ra xkb),
  mm_psim lw G ra phia phib (IEinst xfa xka) (IEinst xfb xkb).
Proof.
  i; rr; splits; vauto; []; exists 0; i; s.
  apply RV in INf; apply RK in INk; des.
  pfold; econs; i.
  mm_destruct XTRa XTRb CNF.
  eapply mm_esim_ok; [by eauto| |].
  - i; invc STb.
    clarify.
    eapply mm_vclosDb in INf1; des; clarify.
    esplits; eauto using IRinst.
  - esplits; eauto; eapply mm_ok_esim_step; eauto; i.
    destruct ma' as [? []]; s in STa; invc STa STb.
    rewrite INf in GET; clarify.
    do 7 eexists; exists true; esplits; first (by eauto); first 1 last; eauto.
    + econs; [by eauto| | |]; first last.
      * inst_pairs; s; noemp; rr; instantiate (5 := pmemp); s; noemp; eauto 10.
      * inst_pairs; s; noemp; rr; instantiate (2 := pmemp); s; noemp; eauto 10.
      * split; rr; eauto.
    + clarify.
      right; econs; eauto.
      * econs; eauto using IRinst.
      * econs 2.
        { rr; eauto. }
        { rr; eauto. }
        { eapply mm_vclosDb in INf1; des; clarify.
          econs 2; rr; eauto. }
        { eapply ksim_mon; eauto using gres_incl_id. }
Qed.


Lemma mm_compat_bop: forall lw G ra phia phib op xa xb ya yb
    (WFx: Ini xb G)
    (WFy: Ini yb G)
    (Ex: xa = match xb with inl xb' => inl (rename_itvar ra xb') | _ => xb end)
    (Ey: ya = match yb with inl yb' => inl (rename_itvar ra yb') | _ => yb end),
  mm_asim lw G ra phia phib (IEbop op xa ya) (IEbop op xb yb).
Proof.
  ii; destruct yb; [apply RV in WFy|]; (destruct xb; [apply RV in WFx|])
  ; (des; s in EV; clarify
     ; try (destruct (lookup (inl i0) envb) eqn:Ei0; rewrite Ei0 in EV; vauto
            ; destruct v; vauto; rewrite Ei0 in INb0; clarify
            ; apply mm_vclosDb in SIM0; des; clarify)
     ; try (destruct (lookup (inl i) envb) eqn:Ei; rewrite Ei in EV; vauto
            ; destruct v; vauto; rewrite Ei in INb; clarify
            ; apply mm_vclosDb in SIM; des; clarify)
     ; esplits; [by s; rewrite ?INa0 ?INa; eauto|]
     ; econs; eapply vclgen_nat; s; eauto).
Qed.


Lemma mm_compat_refeq: forall lw G ra phia phib xa xb ya yb
    (WFx: In (inl xb) G)
    (WFy: In (inl yb) G)
    (Ex: xa = rename_itvar ra xb)
    (Ey: ya = rename_itvar ra yb),
  mm_asim lw G ra phia phib (IErefeq xa ya) (IErefeq xb yb).
Proof.
  ii.
  destruct (RV _ WFx) as [lxa [lxb [Hxa [Hxb Vxab]]]].
  destruct (RV _ WFy) as [lya [lyb [Hya [Hyb Vyab]]]].
  des.
  case_eq lya; move: (mm_vclosDb Vyab); case_eq lyb; ii; subst; des; clarify; try done
  ; rewrite /eval Hyb Hxb in EV; destruct lxb; try done.
  case_eq lxa; move: (mm_vclosDb Vxab); ii; subst; des; clarify.
  simpl; rewrite ?Hxa ?Hya. 
  eexists,,; destruct QR, QR0; des. 
  move: (refst_bij REG0 REG).
  ins; desf; vauto; try by exfalso; eauto.
Qed.


Lemma mm_compat_fst: forall lw G ra phia phib xa xb
    (WF: In (inl xb) G)
    (E: xa = rename_itvar ra xb),
  mm_asim lw G ra phia phib (IEfst xa) (IEfst xb).
Proof.
  ii; apply RV in WF; des.
  destruct (lookup (inl xb) envb) eqn:E'; s in EV; rewrite E' in EV; vauto.
  destruct v; vauto.
  rewrite INb in E'; clarify.
  apply mm_vclosDb in SIM; des; clarify.
  esplits; s; try rewrite INa; eauto.
Qed.


Lemma mm_compat_snd: forall lw G ra phia phib xa xb
    (WF: In (inl xb) G)
    (E: xa = rename_itvar ra xb),
  mm_asim lw G ra phia phib (IEsnd xa) (IEsnd xb).
Proof.
  ii; apply RV in WF; des.
  destruct (lookup (inl xb) envb) eqn:E'; s in EV; rewrite E' in EV; vauto.
  destruct v; vauto.
  rewrite INb in E'; clarify.
  apply mm_vclosDb in SIM; des; clarify.
  esplits; s; try rewrite INa; eauto.
Qed.


Lemma mm_compat_unit: forall lw G ra phia phib,
  mm_asim lw G ra phia phib IEunit IEunit.
Proof.
  ii; s in EV; clarify.
  esplits; s; eauto.
  econs; eapply vclgen_unit; s; eauto.
Qed.


Lemma mm_compat_nat: forall lw G n ra phia phib,
  mm_asim lw G ra phia phib (IEnat n) (IEnat n).
Proof.
  ii; s in EV; clarify.
  esplits; s; eauto.
  econs; eapply vclgen_nat; s; eauto.
Qed.


Lemma mm_compat_pair: forall lw G ra phia phib xa xb ya yb
    (WFx: In (inl xb) G)
    (WFy: In (inl yb) G)
    (Ex: xa = rename_itvar ra xb)
    (Ey: ya = rename_itvar ra yb),
  mm_asim lw G ra phia phib (IEpair xa ya) (IEpair xb yb).
Proof.
  ii; apply RV in WFx; apply RV in WFy; des.
  destruct (lookup (inl xb) envb) eqn:Ex'; s in EV; rewrite Ex' in EV; vauto.
  destruct (lookup (inl yb) envb) eqn:Ey'; s in EV; rewrite Ey' in EV; vauto.
  esplits; s; [by rewrite INa INa0; s; eauto|].
  rewrite Ey' in INb; rewrite Ex' in INb0; clarify.
  econs; eapply vclgen_pair; s; eauto.
Qed.


Lemma mm_compat_inl: forall lw G ra phia phib xa xb
    (WF: In (inl xb) G)
    (E: xa = rename_itvar ra xb),
  mm_asim lw G ra phia phib (IEinl xa) (IEinl xb).
Proof.
  ii; apply RV in WF; des.
  destruct (lookup (inl xb) envb) eqn:Ex; s in EV; rewrite Ex in EV; vauto.
  s in EV; clarify.
  esplits; s; eauto; [by rewrite INa; s; eauto|].
  rewrite Ex in INb; clarify.
  econs; eapply vclgen_inl; s; eauto.
Qed.


Lemma mm_compat_inr: forall lw G ra phia phib xa xb
    (WF: In (inl xb) G)
    (E: xa = rename_itvar ra xb),
  mm_asim lw G ra phia phib (IEinr xa) (IEinr xb).
Proof.
  ii; apply RV in WF; des.
  destruct (lookup (inl xb) envb) eqn:Ex; s in EV; rewrite Ex in EV; vauto.
  s in EV; clarify.
  esplits; s; eauto; [by rewrite INa; s; eauto|].
  rewrite Ex in INb; clarify.
  econs; eapply vclgen_inr; s; eauto.
Qed.


Lemma mm_good_fix: forall (lw: world_loca mm_gw)
    bfa bfb bxa bxb bka bkb ea eb s enva envb
    (GDa: is_good_thunk (IEfix bfa bxa bka ea))
    (GDb: is_good_thunk (IEfix bfb bxb bkb eb)),
  good_thunk (wlift lw) s
             (IVclo enva (IEfix bfa bxa bka ea))
             (IVclo envb (IEfix bfb bxb bkb eb)).
Proof.
  econs 1; s; econs; i; inv FH.
Qed.

Definition purged (phi: clomem) : Prop :=
  forall y e (IN: lookup y phi = Some e), exists x,
    y = inl x.

Lemma respects_purge: forall G env phi
   (RE: respects G env phi)
   (PU: purged phi),
  respects (purge G) env phi.
Proof.
  ii; exploit RE; eauto; []; i; des; subst.
  exploit PU; eauto; []; i; des; subst x.
  esplits; eauto.
  - eauto using purge_subinDl.
  - i. 
    destruct y as [[ly|ty]|ky], x0 as [lx|tx]; simpl in IN1;
    try (eapply EN; s; eapply sublist_in; last eauto; eauto using lbls_upto_sublist');
    try (rewrite /drop_until; s; rewrite drop_until_purge; eapply purge_sublist).
Qed.

Lemma mm_compat_fix: forall lw G ra phia phib bf bx bk ea eb
    (EXf: extendable_t G ra bf bf)
    (EXx: extendable_t G ra bx bx)
    (FRxa: fresh G (inl (inr bx)) phia)
    (FRxb: fresh G (inl (inr bx)) phib)
    (FRfa: fresh G (inl (inr bf)) phia)
    (FRfb: fresh G (inl (inr bf)) phib)
    (FRka: fresh G (inr bk) phia)
    (FRkb: fresh G (inr bk) phib)
    (PUa: purged phia)
    (PUb: purged phib)
    (IH: mm_psim lw (inr bk::inl (inr bx)::inl (inr bf)::purge G)
                 (ra)[inr bf/bf][inr bx/bx][c bk/bk] phia phib ea eb),
  mm_asim lw G ra phia phib (IEfix bf bx bk ea) (IEfix bf bx bk eb).
Proof.
  ii; rrdes IH; s in EV; clarify; esplits; [by s; eauto|].
  econs; eapply vclgen_thunk; econs; s; eauto.
  eapply gk_lsim; eauto; econs; [apply mm_good_fix|].
  { ii; inv FH; eapply GEa; eauto. }
  { ii; inv FH; eapply GEb; eauto. }
  instantiate (1:=wfo_nat (i'+0)).
  pcofix CIH; i.
  apply pesim_gknow; i; apply pesim_gknow_; i.
  ii; invc STK; [|by invc INSTa RF VR].
  invc APPa APPb RF; invc CA CB VR RK.
  eapply pesim_mon; [eapply IH1|by eauto]; eauto; i.
  - repeat (eapply respects_cons); eauto using respects_purge. 
    + split; first eapply FRfa; eauto; ii.
      eapply FRfa; eauto. edestruct RPa; eauto; des.
      rrdes PUa. hexploit PUa; ii; eauto; des; clarify.
      rewrite defbefore_purge in FH. by eapply purge_inD.
    + split; first eapply FRxa; eauto; ii.
      eapply FRxa; eauto. edestruct RPa; eauto; des.
      rrdes PUa. hexploit PUa; ii; eauto; des; clarify.
      rewrite <- defbefore_cons in FH; eauto using purge_subinDl.
      rewrite defbefore_purge in FH. by eapply purge_inD.
    + split; first eapply FRka; eauto; ii.
      eapply FRka; eauto. edestruct RPa; eauto; des.
      rrdes PUa. hexploit PUa; ii; eauto; des; clarify.
      rewrite <- defbefore_cons in FH; s; eauto using purge_subinDl.
      rewrite <- defbefore_cons in FH; s; eauto using purge_subinDl.
      rewrite defbefore_purge in FH. by eapply purge_inD.
  - repeat (eapply respects_cons); eauto using respects_purge. 
    + split; first eapply FRfb; eauto; ii.
      eapply FRfb; eauto. edestruct RPb; eauto; des.
      rrdes PUb. hexploit PUb; ii; eauto; des; clarify.
      rewrite defbefore_purge in FH. by eapply purge_inD.
    + split; first eapply FRxb; eauto; ii.
      eapply FRxb; eauto. edestruct RPb; eauto; des.
      rrdes PUb. hexploit PUb; ii; eauto; des; clarify.
      rewrite <- defbefore_cons in FH; eauto using purge_subinDl.
      rewrite defbefore_purge in FH. by eapply purge_inD.
    + split; first eapply FRkb; eauto; ii.
      eapply FRkb; eauto. edestruct RPb; eauto; des.
      rrdes PUb. hexploit PUb; ii; eauto; des; clarify.
      rewrite <- defbefore_cons in FH; s; eauto using purge_subinDl.
      rewrite <- defbefore_cons in FH; s; eauto using purge_subinDl.
      rewrite defbefore_purge in FH. by eapply purge_inD.
  - eapply bla in EXx; eauto; [].
    eapply foo'; vauto; []; clear IN; i.
    rewrite ren_ext_it_neq_c.
    i; eapply foo'; eauto; vauto.
    { i; clarify; s; tmext_simpl; eauto. }
    { i; clarify.
      destruct (LEM (x0 = inr bx1)); vauto.
      rewrite ren_ext_it_neq in H; [|by ii; vauto].
      destruct x0; vauto.
      exploit (EXx t); s in H; i; des; vauto.
      - apply in_consD in IN; des; vauto.
        eauto using purge_inD.
      - eauto.
    }
    clear IN; i; rewrite ren_ext_it_neq; [eauto|by ii; vauto].
    eapply foo'; eauto; vauto.
    { i; clarify; s; tmext_simpl; splits; eauto.
      econs; eapply vclgen_thunk; econs 1; s; eauto.
      eapply gk_lsim; eauto; econs; [eapply mm_good_fix|].
      { ii; inv FH; eapply GEa; eauto. }
      { ii; inv FH; eapply GEb; eauto. }
      i; eapply gres_incl_id; right; eapply CIH; esplits; eauto using ext_trans.
    }
    { i; clarify.
      destruct (LEM (x1 = inr bf1)); vauto.
      rewrite ren_ext_it_neq in H; [|by ii; vauto].
      destruct x1; vauto.
      exploit (EXf t); s in H; i; des; vauto.
      apply in_consD in IN; des; vauto.
      eauto using purge_inD.
    }
    clear IN; i; rewrite ren_ext_it_neq; [eauto|by ii; vauto].
    hexploit RV; [by eapply purge_inD; eauto|]; i; des; esplits; eauto.
    eapply vclos_mon; eauto; i; eapply gk_mon; eauto.
  - eapply foo'; vauto.
    { i; clarify; s; tmext_simpl; eauto. }
    { i; clarify.
      destruct (LEM (x = bk1)); eauto.
      exfalso.
      do 3 (apply in_consD in IN; des; vauto).
      eauto using purge_inDr.
    }
    { clear IN; i; do 2 (apply in_consD in IN; des; vauto).
      exfalso; eauto using purge_inDr.
    }
Qed.


Lemma mm_good_gen: forall (lw: world_loca mm_gw) bka bkb ea eb s enva envb
    (GDa: is_good_thunk (IEgen bka ea))
    (GDb: is_good_thunk (IEgen bkb eb)),
  good_thunk (wlift lw) s
             (IVclo enva (IEgen bka ea)) (IVclo envb (IEgen bkb eb)).
Proof.
  econs 2; s; econs; i; inv FH.
Qed.

Lemma mm_compat_gen: forall lw G ra phia phib bk ea eb
    (FRka: fresh G (inr bk) phia)
    (FRkb: fresh G (inr bk) phib)
    (PUa: purged phia)
    (PUb: purged phib)
    (IH: mm_psim lw (inr bk :: purge G) (ra)[c bk/bk] phia phib ea eb),
  mm_asim lw G ra phia phib (IEgen bk ea) (IEgen bk eb).
Proof.
  ii; rrdes IH; s in EV; clarify; esplits; [by s; eauto|].
  econs; eapply vclgen_thunk; econs 2; s; eauto.
  eapply gk_lsim; eauto; econs; [apply mm_good_gen|].
  { ii; inv FH; eapply GEa; eauto. }
  { ii; inv FH; eapply GEb; eauto. }
  instantiate (1:=wfo_nat (i'+0)).
  i; apply pesim_gknow; i; apply pesim_gknow_; i.
  ii; invc STK; [by invc APPa RF VR|].
  invc INSTa INSTb RF; invc CA CB VR RK.
  eapply pesim_mon; [eapply IH1|by eauto]; eauto.
  - eapply respects_cons; auto using respects_purge.
    split; first eapply FRka; eauto. ii.
    eapply purge_inDr. eapply sublist_in; first eapply defbefore_sublist; eauto.
  - eapply respects_cons; auto using respects_purge.
    split; first eapply FRkb; eauto. ii.
    eapply purge_inDr. eapply sublist_in; first eapply defbefore_sublist; eauto.
  - i; eapply foo'; eauto; i; vauto.
    hexploit RV; [by eapply purge_inD; eauto|]; i; des; esplits; eauto.
    eapply vclos_mon; eauto; i; eapply gk_mon; eauto.
  - i; eapply foo'; eauto; vauto. 
    { i; clarify; s; tmext_simpl; eauto. }
    { i; clarify.
      destruct (LEM (x = bk1)); eauto.
      exfalso.
      apply in_consD in IN; des; vauto.
      eauto using purge_inDr.
    }
    { clear IN; i.
      exfalso; eauto using purge_inDr.
    }
Qed.


Lemma mm_compat_let: forall lw G ra phia phib ea ea' bx eb eb'
    (EX: extendable_t G ra bx bx)
    (FRa: fresh G (inl (inr bx)) phia)
    (FRb: fresh G (inl (inr bx)) phib)
    (IH: mm_asim lw G ra phia phib ea eb)
    (IH': mm_psim lw (inl (inr bx) :: G) ra[inr bx/bx] phia phib ea' eb'),
  mm_psim lw G ra phia phib (IElet ea (inl bx) ea') (IElet eb (inl bx) eb').
Proof.
  i; rr; splits; vauto; []; exists 0; i; s.
  rrdes IH IH'; pfold; econs; i.
  mm_destruct XTRa XTRb CNF.
  eapply mm_esim_ok; [by eauto| |].
  - i; invc STb.
    hexploit IH; eauto; i; des; eauto using IRlet.
  - destruct (LEM (exists vb, eval envb eb = Some vb)); last first.
    { i; eapply mm_ok_esim_err.
        by econs; eapply IRerr; eauto; i; inv FH; vauto; apply f; vauto.
      s; rr; i; inv FH; des; clarify; rrdes PV; vauto. }
    des; esplits; eauto; eapply mm_ok_esim_step; eauto; i.
    destruct ma' as [? []]; s in STa; invc STa.
    do 7 eexists; exists false; splits; first (by eauto); last (by s; eauto)
    ; first 1 last.
    + econs; eauto; noemp; inst_pairs; s; noemp; rr; eauto 10.
    + left; econs; eauto.
      * left; eapply IRok; econs; eauto.
      * eapply gres_incl_id; left.
        rewrite 2!mid_csplit.
        eapply pesim_release; eauto; try (by rr; eauto).
        eapply IH'1; eauto using respects_cons; i.
        { eapply foo'; eauto.
          { i; clarify; s rename_itvar; rewrite tmext_eq; splits; eauto.
            exploit IH; eauto; i; des.
            rewrite x4 in EVAL; clarify.
          }
          { i; clarify.
            destruct (LEM (x2 = inr bx)); vauto.
            rewrite ren_ext_it_neq in H; [|by ii; vauto].
            destruct x2; vauto.
            exploit (EX t); s in H; i; des; vauto.
            apply in_consD in IN; des; vauto.
          } 
          clear IN; i; rewrite ren_ext_it_neq; [|by ii; vauto].
          apply RV in IN; des; esplits; eauto.
        }
        { apply in_consD in IN; des; vauto.
          rewrite ren_ext_c_neq.
          rewrite !list_assoc_cons_neq; eauto; vauto.
        }
Qed.


Lemma mm_compat_cont: forall lw G ra phia phib bx ea eb
    (EX: extendable_t G ra bx bx)
    (FRa: fresh G (inl (inr bx)) phia)
    (FRb: fresh G (inl (inr bx)) phib)
    (IH: mm_psim lw (inl (inr bx) :: G) ra[inr bx/bx] phia phib ea eb),
  mm_ksim lw G ra phia phib (IEcont bx ea) (IEcont bx eb).
Proof.
  ii; clarify; rrdes IH; esplits; eauto.
  econs; i; invc RETa RETb.
  eapply pesim_gknow; i.
  pfold; econs; i.
  mm_destruct XTRa XTRb CNF.
  esplits; eauto; eapply esim_step; eauto; i.
  invc ST; [invc STEP|by exfalso; eapply STK; econs; eauto].
  do 7 eexists; exists false; splits; first (by eauto); last (by s; eauto)
  ; last (econs; [by eauto|by eauto| |]); last first; first 1 last.
  - inst_pairs; s; noemp; rr; eauto 10.
  - left; econs; eauto.
    + left; inst_pairs; eapply IRok; econs; eauto.
    + rewrite GET in GETk; rewrite GETV in GETv; clarify.
      eapply gres_incl_id; left.
      rewrite mid_csplit; instantiate (1:=_#_).
      clear CO CO0.
      eapply pesim_release; eauto; try (by rr; eauto).
      inst_pairs; eapply IH1; eauto using respects_cons, extpub_trans.
      * i; eapply foo'; eauto.
        { i; clarify; s rename_itvar; rewrite tmext_eq; splits; eauto. }
        { i; clarify.
          destruct (LEM (x2 = inr bxb)); vauto.
          rewrite ren_ext_it_neq in H; [|by ii; vauto].
          destruct x2; vauto.
          exploit (EX t); s in H; i; des; vauto.
          apply in_consD in IN; des; vauto.
        } 
        clear IN; i; rewrite ren_ext_it_neq; [|by ii; vauto].
        apply RV in IN; des; esplits; eauto.
        eapply vclos_mon; eauto; i; eapply gk_mon; eauto.
      * i; eapply foo'; i; clarify; eauto.
        eapply RK in IN0; des; esplits; eauto.
        eapply ksim_mon; eauto.
  - rewrite GET in GETk; rewrite GETV in GETv; clarify.
    s; noemp; rr; eauto 10.
Qed.


Lemma mm_compat_letcont: forall lw G ra phia phib ea ea' bk eb eb'
    (EX: extendable_c G ra bk bk)
    (FRa: fresh G (inr bk) phia)
    (FRb: fresh G (inr bk) phib)
    (IH: mm_ksim lw G ra phia phib ea eb)
    (IH': mm_psim lw (inr bk :: G) ra[c bk/bk] phia phib ea' eb'),
  mm_psim lw G ra phia phib (IElet ea (inr bk) ea') (IElet eb (inr bk) eb').
Proof.
  i; rr; splits; vauto; []; exists 0; i; s.
  rrdes IH IH'; pfold; econs; i.
  mm_destruct XTRa XTRb CNF.
  
  
  
  
  
  
  
  
  
  
  destruct (LEM (forall evt hb'' envb' e',
                   ~ reduce_ok evt ((hb' ## x3) ## x0) envb (IElet eb (inr bk) eb')
                                   hb'' envb' e')).
  { eapply esim_err.
      by econs; eapply IRerr; eauto.
    rr; i; inv FH; des; clarify; rrdes PV; vauto. } 
  repeat (eapply not_all_ex_not in f; des); clear f; inv n3.
  hexploit IH; eauto; i; des; clarify.
  esplits; eauto; eapply esim_step; eauto; i.
  invc ST; [invc STEP|by exfalso; eapply STK; eauto].
  do 7 eexists; exists false; splits; first (by eauto); last (by s; eauto)
  ; last (econs; [by eauto|by eauto| |]); last first; first 1 last.
  - inst_pairs; s; noemp; rr; eauto 10.
  - left; econs; eauto.
    + left; inst_pairs; eapply IRok; econs; eauto.
    + eapply gres_incl_id; left.
      rewrite mid_csplit; instantiate (1:=_#_).
      clear CO CO0.
      eapply pesim_release; eauto; try (by rr; eauto).
      inst_pairs; eapply IH'1; eauto using respects_cons, extpub_trans.
      * 
        i; eapply foo'; i; clarify; eauto.
      * i; eapply foo'; eauto.
        { i; clarify; simpl rename_contvar; s; tmext_simpl; eauto. }
        { i; clarify.
          destruct (LEM (x2 = bk)); eauto.
          rewrite tmext_neq in H; [|by ii; vauto].
          exploit (EX x2); s in H; i; des; vauto.
          apply in_consD in IN; des; vauto.
        } 
        clear IN; i.
        rewrite ren_ext_c_neq'; [|by ii; vauto].
        eapply RK in IN; des; esplits; eauto.
  - s; noemp; rr; eauto 10.
Qed.

Lemma extendable_c_same: forall G ren x
   (E: forall y (IN: In y G) (E: rename_var ren y = inr x), rename_var ren y = y),
 extendable_c G ren x x.
Proof.
 s; ii; clarify; hexploit (E (inr xb')); eauto.
 s; i; clarify; rewrite H in E0; clarify.
Qed.



Hint Resolve 
  mm_compat_input mm_compat_output mm_compat_app mm_compat_read
  mm_compat_write mm_compat_new mm_compat_ret mm_compat_ifnz
  mm_compat_case mm_compat_inst mm_compat_bop mm_compat_refeq
  mm_compat_fst mm_compat_snd mm_compat_unit mm_compat_nat
  mm_compat_pair mm_compat_inl mm_compat_inr mm_compat_fix
  mm_compat_gen mm_compat_let mm_compat_cont mm_compat_letcont: mm_test_compat.




























































































































































































Require mm.


Lemma mm_psim_phi_weaken : forall lw ea eb G phia phib
    (SIM: mm.mm_psim lw G ren_id ea eb),
  mm_psim lw G ren_id phia phib ea eb.
Proof.
  ii; rrdes SIM; rr; splits; i; auto.
  exists i'; i; apply SIM1; auto.
Qed.

Lemma mm_asim_phi_weaken : forall lw ea eb G phia phib
    (SIM: mm.mm_asim lw G ren_id ea eb),
  mm_asim lw G ren_id phia phib ea eb.
Proof.
  ii; rrdes SIM.
  rr; splits; i; eauto.
Qed.

Lemma mm_ksim_phi_weaken : forall lw ea eb G phia phib
    (SIM: mm.mm_ksim lw G ren_id ea eb),
  mm_ksim lw G ren_id phia phib ea eb.
Proof.
  ii; rrdes SIM; eapply SIM; eauto.
Qed.



















Inductive mm_tlsim lw FI phia phib : exp -> exp -> Prop :=
  | mm_tlsim_fix bf bx bk ea eb
      (SIM: mm_psim lw (inr bk::inl (inr bx)::inl (inr bf)::map (inl<*>inl) FI)
                    ren_id phia phib ea eb)
    : mm_tlsim lw FI phia phib (IEfix bf bx bk ea) (IEfix bf bx bk eb)
  | mm_tlsim_gen bk ea eb
      (SIM: mm_psim lw (inr bk::map (inl<*>inl) FI)
                    ren_id phia phib ea eb)
    : mm_tlsim lw FI phia phib (IEgen bk ea) (IEgen bk eb)
.

Lemma mm_tlsim_tlsim: forall lw suba subb s phia phib ea eb
    (VSIM: mm_tlsim lw (map fst suba) phia phib ea eb)
    (DOM: map fst suba = map fst subb)
    (REa: respects (map (inl <*> inl) (map fst suba))
                   (map_pair (inl <*> inl) id suba) phia)
    (REb: respects (map (inl <*> inl) (map fst subb))
                   (map_pair (inl <*> inl) id subb) phib)
    (LBLa: forall x e (IN: lookup x phia = Some e),
             exists l, x = inl (inl l))
    (LBLb: forall x e (IN: lookup x phib = Some e),
             exists l, x = inl (inl l)),
  @tlsim _ _ mm_gw lw (map fst suba) suba subb s
         (IVclo (map_pair (inl <*> inl) id suba) ea)
         (IVclo (map_pair (inl <*> inl) id subb) eb).
Proof.
  i; invc VSIM.
  - econs 1; s; eauto.
    dup SIM; rrdes SIM0.
    rewrite map_comp -purge_map_inl -map_comp in SIM.

    replace ren_id with (((ren_id)[inr bf/bf])[inr bx/bx])[c bk/bk] in SIM; [|]; last first.
    { by rewrite !ren_id_self_t !ren_id_self_c. }
    eapply mm_compat_fix in SIM; eauto; first 1 last;
     try ( ii; try (apply LBLa in IN || apply LBLb in IN); des; subst; splits; vauto; ii; 
           eapply sublist_in in FH; last eapply defbefore_sublist;
           eapply in_map_iff in FH; des; clarify );
     try ( ii; (apply LBLa in IN || apply LBLb in IN); des; subst; esplits; eauto ).
    { eapply extendable_t_same; i; rewrite ren_var_id; auto. }
    { eapply extendable_t_same; i; rewrite ren_var_id; auto. }
    





    eapply gk_lsim; eauto.
    econs; [eapply mm_good_fix; eauto|].
    { ii; inv FH; eapply GEa; eauto. }
    { ii; inv FH; eapply GEb; eauto. }
    i; invc STK; [|by invc RF VR INSTa CA].
    invc RK RF VR; invc APPa APPb CA CB.
    eapply pesim_gknow; i.
    eapply SIM0; eauto.
    + apply respects_cons; auto;
      first apply respects_cons; auto;
      first apply respects_cons; auto;
      ii; hexploit LBLa; ii; eauto; des; subst; s; splits; vauto; ii;
      eapply sublist_in in FH; try eapply lbls_upto_sublist;
      eapply in_map_iff in FH; des; clarify.
    + apply respects_cons; auto;
      first apply respects_cons; auto;
      first apply respects_cons; auto;
      rewrite DOM; eauto;
      ii; hexploit LBLb; ii; eauto; des; subst; s; splits; vauto; ii;
      eapply sublist_in in FH; try eapply lbls_upto_sublist;
      eapply in_map_iff in FH; des; clarify.
    + i; apply in_consD in IN; des; [by vauto|].
      rewrite (list_assoc_cons_neq _ ka); [|by vauto].
      rewrite (list_assoc_cons_neq _ kb); [|by vauto].
      apply in_consD in IN; des.
      { clarify; rewrite 2!list_assoc_cons_eq; esplits; eauto. }
      rewrite ren_itvar_id.
      rewrite (list_assoc_cons_neq _ va); [|by ii; vauto].
      rewrite (list_assoc_cons_neq _ vb); [|by ii; vauto].
      apply in_consD in IN; des.
      { clarify; rewrite 2!list_assoc_cons_eq; esplits; eauto.
        exploit SIM; try (by s; eauto).
        - rewrite DOM; eauto.
        - 
          
          i; apply in_map_iff in IN; des; s in IN; clarify.
          dup IN2; dup IN2; rewrite DOM in IN2.
          eapply list_assoc_map in IN2; eapply list_assoc_map in IN3; des.
          eapply RS in IN4; eauto; esplits.
          + eapply list_assoc_map_pair; eauto; i; ss; clarify.
          + eapply list_assoc_map_pair; eauto; i; ss; clarify.
          + econs; eapply vclgen_thunk, vclos_thunk_mon; eauto.
            i; eapply LE0, gk_mon; eauto.
        - i; des; s in x0; clarify.
      }
      rewrite list_assoc_cons_neq; [|by ii; vauto].
      rewrite list_assoc_cons_neq; [|by ii; vauto].
      eapply in_map_iff in IN; des; s in IN; clarify.
      hexploit list_assoc_map; eauto; []; i; des.
      rewrite DOM in IN3.
      hexploit list_assoc_map; eauto; []; i; des.
      rewrite -DOM in IN3.
      hexploit RS; eauto.
      i; esplits.
      * eapply list_assoc_map_pair; eauto; i; ss; clarify.
      * eapply list_assoc_map_pair; eauto; i; ss; clarify.
      * econs; eapply vclgen_thunk, vclos_thunk_mon; eauto.
        i; eapply LE0, gk_mon; eauto.
    + i; apply in_consD in IN; des.
      { clarify; rewrite 2!list_assoc_cons_eq; esplits; eauto. }
      rewrite (list_assoc_cons_neq _ ka); [|by ii; vauto].
      rewrite (list_assoc_cons_neq _ kb); [|by ii; vauto].
      apply in_consD in IN; des; [by vauto|].
      rewrite (list_assoc_cons_neq _ va); [|by ii; vauto].
      rewrite (list_assoc_cons_neq _ vb); [|by ii; vauto].
      apply in_consD in IN; des; [by vauto|].
      exfalso; apply in_map_iff in IN; des; vauto.
  - 
    econs 2; s; eauto.
    dup SIM; rrdes SIM0.
    
    eapply gk_lsim; eauto.
    econs; [eapply mm_good_gen; eauto|].
    { ii; inv FH; eapply GEa; eauto. }
    { ii; inv FH; eapply GEb; eauto. }
    i; invc STK; [by invc RF VR APPa CA|].
    invc RK RF VR; invc INSTa INSTb CA CB.
    eapply pesim_gknow; i.
    eapply SIM0; eauto.
    + apply respects_cons; auto. 
      ii; hexploit LBLa; ii; eauto; des; subst; s; splits; vauto; ii;
      eapply sublist_in in FH; try eapply lbls_upto_sublist;
      eapply in_map_iff in FH; des; clarify.
    + apply respects_cons; auto;
      rewrite DOM; eauto;
      ii; hexploit LBLb; ii; eauto; des; subst; s; splits; vauto; ii;
      eapply sublist_in in FH; try eapply lbls_upto_sublist;
      eapply in_map_iff in FH; des; clarify.
    + i; apply in_consD in IN; des; [by vauto|].
      rewrite (list_assoc_cons_neq _ ka); [|by vauto].
      rewrite (list_assoc_cons_neq _ kb); [|by vauto].
      eapply in_map_iff in IN; des; s in IN; clarify.
      hexploit list_assoc_map; eauto; []; i; des.
      rewrite DOM in IN1.
      hexploit list_assoc_map; eauto; []; i; des.
      rewrite -DOM in IN1.
      hexploit RS; eauto.
      i; esplits.
      * eapply list_assoc_map_pair; eauto; i; ss; clarify.
      * eapply list_assoc_map_pair; eauto; i; ss; clarify.
      * econs; eapply vclgen_thunk, vclos_thunk_mon; eauto.
        i; eapply LE0, gk_mon; eauto.
    + i; apply in_consD in IN; des.
      { clarify; rewrite 2!list_assoc_cons_eq; esplits; eauto. }
      rewrite (list_assoc_cons_neq _ ka); [|by ii; vauto].
      rewrite (list_assoc_cons_neq _ kb); [|by ii; vauto].
      exfalso; apply in_map_iff in IN; des; vauto.
Grab Existential Variables.
  exact 0. exact 0.
Qed.


























Definition mm_msim ienv oenv (phia phib: clomem) moda modb : Prop :=
  exists (lw: world_loca mm_gw),
  forall ancha suba ciga cila (LMa: mid_lsi.(cload) moda ancha suba ciga cila)
         anchb subb cigb cilb (LMb: mid_lsi.(cload) modb anchb subb cigb cilb)
         (SUBa: map fst suba = ienv)
         (SUBb: map fst subb = ienv)
         (PHIa: respects (map (inl<*>inl) ienv) (map_pair (inl<*>inl) id suba) phia)
         (PHIb: respects (map (inl<*>inl) ienv) (map_pair (inl<*>inl) id subb) phib),
  exists si,
      <<STA: stable lw>> /\
      <<INIg: forall g (GK: gknow (wlift lw) g), mm_gw.(confrel) (fun sg => g (sg, snd si)) (fst si) ciga cigb>> /\
      <<INIl: forall g (GK: gknow (wlift lw) g), lw.(confrel) g si cila cilb>> /\
      <<REF: forall la lb, ~ mm_gw.(rquery) (fst si) la lb>> /\
      <<SIM: forall f (IN: In f oenv),
               exists va vb,
                 <<Ea: vload _ moda ancha suba f va>> /\
                 <<Eb: vload _ modb anchb subb f vb>> /\
                 <<SIM: tlsim ienv suba subb si va vb>> >>.

Lemma respects_nil: forall G env,
  respects G env nil.
Proof.
  ii; vauto.
Qed.
Hint Resolve respects_nil.

Lemma mm_msim_msim: forall FI FO ma mb
    (SIM: mm_msim FI FO nil nil ma mb),
  msim mm_gw FI FO ma mb.
Proof.
  i; rrdes SIM; rr; eauto.
Qed.


Lemma mm_msim_nil: forall FI phia phib,
  mm_msim FI nil phia phib nil nil.
Proof.
  i; exists (lwemp mm_gw); i; exists (tt, tt, refst_emp, tt).
  invc LMa LMb; clean.
  esplits; eauto using lwemp_stable.
  - i; rr; repeat ex @cemp; splits.
    + split; rr; auto.
    + apply mm_refconfrel_emp; eauto.
    + s; noemp; auto.
    + s; noemp; auto.
  - i; rr; eauto.
  - i; rrdes FH; inv REG.
  - i; vauto.
Qed.

Lemma lbls_upto_last y z (xs : list var):
  In z (lbls_upto y (xs ++ [inl (inl y)])) ->
  In z xs.
Proof.
  ii. subst. induction xs; eauto. 
  - s in H. destruct eq_nat_dec; clarify.
  - destruct a as [[la|ta]|ka]; simpl in *.
    + destruct eq_nat_dec; clarify.
      apply in_consD in H; des; clarify; auto.
    + right. by eapply IHxs.
    + right. by eapply IHxs.
Qed.

Lemma lbls_upto_prefix x y z (xs : list var):
  In (inl (inl x)) xs ->
  In z (lbls_upto x (xs ++ [inl (inl y)])) ->
  In z xs.
Proof.
  ii. subst. induction xs; eauto.
  apply in_consD in H; des; clarify.
  - s in H0. destruct eq_nat_dec; clarify.
  - destruct a as [[la|ta]|ka]; s; auto.
    s in H0. destruct eq_nat_dec; clarify.
      apply in_consD in H0; des; clarify; auto.
Qed.

Lemma lbls_upto_prefix' x y z (xs : list var):
  In (inl (inl x)) xs ->
  In z (lbls_upto x (xs ++ [inl (inl y)])) ->
  In z (lbls_upto x xs).
Proof.
  ii. subst. induction xs; eauto.
  apply in_consD in H; des; clarify.
  - s in H0. destruct eq_nat_dec; clarify.
  - destruct a as [[la|ta]|ka]; s; auto.
    s in H0. destruct eq_nat_dec; s; clarify.
      apply in_consD in H0; des; clarify; auto.
Qed.

Lemma mm_msim_cons: forall lw FI ea eb l FO ma mb sil phia phib
    (VSIM: mm_tlsim lw FI phia phib ea eb)
    (MSIM: mm_msim (FI++[l]) FO 
                   ((inl (inl l), ea) :: phia) ((inl (inl l), eb) :: phib)
                   ma mb)
    (FR: ~ In l FI)
    
    
    
    
    
    
    (LBLa: forall x e (IN: lookup x phia = Some e),
             exists l, x = inl (inl l))
    (LBLb: forall x e (IN: lookup x phib = Some e),
             exists l, x = inl (inl l))
    (ST: stable lw)
    (INIl: forall g (GK: gknow (wlift lw) g),
             lw.(confrel) g (tt, tt, refst_emp, sil) cemp cemp),
  mm_msim FI (l::FO) phia phib ((l,ea)::ma) ((l,eb)::mb).
Proof.
  i; rrdes MSIM.
  exists (wprod lw lw0); i.
  set (va := IVclo (map_pair (inl <*> inl) id suba) ea).
  set (vb := IVclo (map_pair (inl <*> inl) id subb) eb).
  hexploit MSIM; try by rr; eauto.
  { instantiate (1 := suba ++ [(l, va)]).
    rewrite map_app SUBa; auto.
  }
  { instantiate (1 := subb ++ [(l, vb)]).
    rewrite map_app SUBb; auto.
  }
  { rewrite map_app (map_pair_app eqdec); s. 
    ii; destruct (LEM (x = inl (inl l))).
    - subst. rewrite list_assoc_app.
      destruct (@list_assoc var val eqdec (inl (inl l)) (map_pair (inl <*> inl) id suba)) eqn:E.
      { exfalso. apply list_assoc_someD in E.
        unfold map_pair in E; apply in_map_iff in E; des; clarify.
        apply FR; eauto using in_map.
      }
      rewrite list_assoc_cons_eq; rewrite list_assoc_cons_eq in IN; clarify.
      subst e; unfold va; rewrite E; esplits.
      + eapply in_or_app; s; eauto.
      + instantiate (1:= (map_pair (inl <*> inl) id suba)); eauto. 
      + subst. i.
        rewrite list_assoc_app.
        case_eq (lookup y (map_pair (inl <*> inl) id suba)).
        * ii. by rewrite H.
        * ii. rewrite H.
          apply lbls_upto_last in IN.
          apply in_map_iff in IN; des; subst.
          edestruct list_assoc_map; eauto.
          erewrite list_assoc_map_pair in H; eauto; ii; clarify.
    - rewrite list_assoc_cons_neq in IN; [|by ii; vauto].
      exploit PHIa; eauto; []; i; des.
      rewrite list_assoc_app.
      esplits.
      + eauto using in_or_app.
      + rewrite IN0. eauto.
      + subst. i. rewrite list_assoc_app.
        eapply in_map_iff in IG; des; clarify.
        s in IN1; dup IN1; apply lbls_upto_prefix' in IN1; 
        last by (eapply in_map_iff; eexists; s; eauto).
        apply EN in IN1.
        rewrite IN1.
        case_eq (lookup y (map_pair (inl <*> inl) id suba)).
        * ii. by rewrite H.
        * ii. rewrite H.
          apply lbls_upto_prefix in IN2;
          last by (eapply in_map_iff; eexists; s; eauto).
          apply in_map_iff in IN2; des; subst.
          edestruct list_assoc_map; eauto.
          erewrite list_assoc_map_pair in H; eauto; ii; clarify.
  }
  { rewrite map_app (map_pair_app eqdec); s. 
    ii; destruct (LEM (x = inl (inl l))).
    - subst. rewrite list_assoc_app.
      destruct (@list_assoc var val eqdec (inl (inl l)) (map_pair (inl <*> inl) id subb)) eqn:E.
      { exfalso. apply list_assoc_someD in E.
        unfold map_pair in E; apply in_map_iff in E; des; clarify.
        apply FR; eauto using in_map. rewrite <- SUBb.
        by eapply in_map.
      }
      rewrite list_assoc_cons_eq; rewrite list_assoc_cons_eq in IN; clarify.
      subst e; unfold va; rewrite E; esplits.
      + eapply in_or_app; s; eauto.
      + instantiate (1:= (map_pair (inl <*> inl) id subb)); eauto. 
      + subst. i.
        rewrite list_assoc_app.
        case_eq (lookup y (map_pair (inl <*> inl) id subb)).
        * ii. by rewrite H.
        * ii. rewrite H.
          apply lbls_upto_last in IN.
          apply in_map_iff in IN; des; subst.
          edestruct list_assoc_map; eauto.
          eapply (list_assoc_eqdom _ _ _ (eq_sym SUBb)) in H0; des.
          erewrite list_assoc_map_pair in H; eauto; ii; clarify.
    - rewrite list_assoc_cons_neq in IN; [|by ii; vauto].
      exploit PHIb; eauto; []; i; des.
      rewrite list_assoc_app.
      esplits.
      + eauto using in_or_app.
      + rewrite IN0. eauto.
      + subst. i. rewrite list_assoc_app.
        eapply in_map_iff in IG; des; clarify.
        s in IN1; dup IN1; apply lbls_upto_prefix' in IN1; 
        last by (eapply in_map_iff; eexists; s; eauto).
        apply EN in IN1.
        rewrite IN1.
        case_eq (lookup y (map_pair (inl <*> inl) id subb)).
        * ii. by rewrite H.
        * ii. rewrite H.
          apply lbls_upto_prefix in IN2;
          last by (eapply in_map_iff; eexists; s; eauto).
          apply in_map_iff in IN2; des; subst.
          edestruct list_assoc_map; eauto.
          eapply (list_assoc_eqdom _ _ _ (eq_sym SUBb)) in H0; des.
          erewrite list_assoc_map_pair in H; eauto; ii; clarify.
  }
  i; des.
  exists (fst si, (sil, snd si)).
  rrdes LMa LMb; subst.
  assert (fst si = (tt, tt, refst_emp)).
  { destruct si as [[[[][]]]]; s; repeat f_equal.
    hexploit INIg; eauto using gkmin_bot_gknow; []; i.
    rrdes X WHX.
    destruct hax, hay, hbx, hby; s in HPA; s in HPB; clarify.
    symmetry in H1; apply pmemp_plus in H1.
    symmetry in H3; apply pmemp_plus in H3.
    des; subst; rrdes WHY; clarify.
    apply refst_extensionality; split; i; vauto.
    apply RELV in H; des; apply pmread_empD in RDa; vauto.
  }
  destruct si; clean; subst.

  assert (SIM': @tlsim _ _ mm_gw lw (map fst suba) suba subb
                       (tt, tt, refst_emp, sil) va vb).
  { rewrite -SUBb in PHIb; eapply mm_tlsim_tlsim; eauto. }

  splits; eauto.
  - eapply stable_prod; eauto.
  - i; hexploit gknow_prod; first 2 last; eauto; []; i; des.
    eapply confrel_mon; [by eapply INIg; eauto|].
    i; rrdes PR; inv PR; eapply gk_mon; s; eauto.
  - i; hexploit gknow_prod; first 2 last; eauto; []; i; des.
    move INIl at bottom; move INIl0 at bottom.
    rr; esplits; clean; eauto; noemp; eauto.
  - i; apply in_consD in IN; des; subst.
    + esplits; try (rr; s; des_if); vauto.
      ii; rrdes LE LE0; destruct s as [[[[]]]]; destruct s2; clean.
      hexploit gknow_prod; first 2 last; eauto; []; i; des.
      exploit SIM'.
      * inst_pairs; split; [exact LE|exact LE0].
      * eauto.
      * rrdes RLS RLS; rr; eauto.
      * i; exploit RS; eauto; clean; i.
        eapply vclos_thunk_lift, vclos_thunk_unlift, vclos_thunk_mon; eauto.
      * i; eapply vclos_thunk_lift, vclos_thunk_unlift, vclos_thunk_mon; eauto.
        i; invc PR; eapply GK; repeat (eauto; split; clean).
    + hexploit SIM; eauto; []; i; des.
      esplits.
      * rr; s; des_if; [by exfalso; vauto|].
        rr in Ea; eauto.
      * rr; s; des_if; [by exfalso; vauto|].
        rr in Eb; eauto.
      * { ii; rrdes LE LE0; destruct s as [[[]]]; destruct s3; clean.
          i; hexploit gknow_prod; first 2 last; eauto; []; i; des.
          exploit SIM0.
          - inst_pairs; split; [exact LE|exact LE1].
          - eauto.
          - rrdes RLS RLS; rr; eauto.
          - i; eapply (lookup2 _ _ _ _ _ _ _ SUBb); eauto; i.
            + exploit RS; eauto.
              * apply in_map_fst; eauto using list_assoc_someD.
              * clean; i.
                eapply vclos_thunk_lift; eapply vclos_thunk_unlift in x1.
                eapply vclos_thunk_mon; eauto.
            + sng; exploit SIM'.
              * inst_pairs; split; [exact LE|exact LE0].
              * eauto.
              * rrdes RLS RLS; rr; esplits; rr; clean; esplits; eauto.
              * i; exploit RS; eauto; []; i.
                eapply vclos_thunk_lift, vclos_thunk_unlift.
                eapply vclos_thunk_mon; eauto.
              * i; eapply vclos_thunk_lift; eapply vclos_thunk_unlift in x0.
                eapply vclos_thunk_mon; eauto.
                i; inv PR; econs.
                { eapply GK; repeat (eauto; split; clean). }
                eauto.
          - i; eapply vclos_thunk_lift, vclos_thunk_unlift.
            eapply vclos_thunk_mon; eauto.
            i; invc PR; eapply GK; repeat (eauto; split; clean).
        }
Grab Existential Variables.
  exact tt. exact tt.
Qed.




































































































































































































































































































































































































































































































































            
























































































































